# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test functionality for the PathEnsemble classes."""
from io import StringIO
import logging
import os
import numpy as np
import tempfile
import unittest
from pyretis.core.particles import Particles, ParticlesExt
from pyretis.core.path import Path
from pyretis.core.pathensemble import (
    _generate_file_names,
    get_path_ensemble_class,
    PathEnsemble,
    PathEnsembleExt,
    generate_ensemble_name,
)
from pyretis.core.random_gen import (MockRandomGenerator,
                                     create_random_generator)
from pyretis.core.system import System
from pyretis.inout.common import make_dirs
from unittest.mock import patch
from .help import PATHTEST0, PATHTEST1, turn_on_logging
logging.disable(logging.CRITICAL)


FILE_NAME = 'file{:03d}.xyz'
DIR_NAME = os.path.abspath(os.path.join('path', 'to'))
HERE = os.path.abspath(os.path.dirname(__file__))
DIRS = [
    os.path.join(HERE, '001'),
    os.path.join(HERE, '001', 'accepted'),
    os.path.join(HERE, '001', 'accepted', 'not-needed'),
    os.path.join(HERE, '001', 'generate'),
    os.path.join(HERE, '001', 'traj'),
]


def create_dirs():
    """Just create some dirs we need for testing."""
    for i in DIRS:
        make_dirs(i)


def remove_dirs():
    """Remove the dirs we created."""
    for i in DIRS:
        try:
            os.removedirs(i)
        except OSError:
            pass


def make_system(order, pos, vel, vpot, ekin, internal=True):
    """Create a system for testing."""
    system = System()
    if internal:
        system.particles = Particles(dim=3)
    else:
        system.particles = ParticlesExt(dim=3)
    system.order = order
    system.particles.set_pos(pos)
    system.particles.set_vel(vel)
    system.particles.vpot = vpot
    system.particles.ekin = ekin
    return system


def make_fake_extpath(length=10):
    """Just return a fake path for testing."""
    rgen = MockRandomGenerator(seed=0)
    path = Path(rgen)
    for i in range(length):
        filename = os.path.join(DIR_NAME, FILE_NAME.format(i))
        path.append(
            make_system([i], (filename, i), False, 0.0, 0.0, internal=False)
        )
    return path


def make_fake_path(length=10):
    """Just return a fake path for testing."""
    rgen = MockRandomGenerator(seed=0)
    path = Path(rgen)
    for i in range(length):
        path.append(
            make_system([i], np.ones((5, 3)) * i,
                        np.ones((5, 3)) * i, 0.0, 0.0)
        )
    path.generated = ('fake',)
    return path


def make_fake_path_files(dirname):
    """Return a fake path and generate files for it."""
    rgen = MockRandomGenerator(seed=0)
    path = Path(rgen)
    for name in ('fake_path_1', 'fake_path_2'):
        filename = os.path.join(dirname, name)
        with open(filename, 'w', encoding='utf-8') as output:
            output.write('Ibsens ripsbusker og andre buskevekster\n')
            output.write(filename)
    for i in range(10):
        if i < 5:
            filename = os.path.join(dirname, 'fake_path_1')
        else:
            filename = os.path.join(dirname, 'fake_path_2')
        path.append(
            make_system([i], (filename, i), False, 0.0, 0.0, internal=False)
        )
    path.generated = ('fake',)
    return path


def remove_path_files(path):
    """Remove the files for a trajectory."""
    # Just remove the files:
    names = set()
    for i in path.phasepoints:
        names.add(i.particles.get_pos()[0])
    for name in names:
        _remove_file(name)


def _remove_file(name):
    """Silently remove file."""
    try:
        os.remove(name)
    except OSError:
        pass


class MethodsTest(unittest.TestCase):
    """Run the tests for the PathEnsemble class."""

    def test_generate_file_name(self):
        """Test the generation of file names."""
        path = make_fake_extpath()
        new_dir = os.path.abspath(os.path.join('new', 'target'))
        new_pos, source = _generate_file_names(path, new_dir)
        for i, (point, pointn) in enumerate(zip(path.phasepoints, new_pos)):
            self.assertEqual(i, pointn[1])
            self.assertEqual(point.particles.get_pos()[1], pointn[1])
            path1 = os.path.dirname(point.particles.get_pos()[0])
            path2 = os.path.dirname(pointn[0])
            self.assertEqual(path1, DIR_NAME)
            self.assertEqual(path2, new_dir)
            self.assertTrue(point.particles.get_pos()[0] in source)
            target = source[point.particles.get_pos()[0]]
            self.assertEqual(target, pointn[0])
        new_pos2, _ = _generate_file_names(path, new_dir, prefix='prefix-')
        for i, point in enumerate(new_pos2):
            self.assertEqual(
                os.path.basename(point[0]),
                f'prefix-{FILE_NAME.format(i)}',
            )

    def test_get_class(self):
        """Test that we get the correct class."""
        klass1 = get_path_ensemble_class('internal')
        self.assertTrue(klass1 is PathEnsemble)
        klass2 = get_path_ensemble_class('external')
        self.assertTrue(klass2 is PathEnsembleExt)
        with self.assertRaises(ValueError):
            get_path_ensemble_class('Pretty fly for a whity guy')

    def test_generate_ensemble_name(self):
        """Test that we can generate names for directories."""
        name = generate_ensemble_name(5, zero_pad=3)
        self.assertEqual(name, '005')
        name = generate_ensemble_name(2, zero_pad=6)
        self.assertEqual(name, '000002')
        for i in (-1, 0, 1, 2):
            name = generate_ensemble_name(1, zero_pad=i)
            self.assertEqual(name, '001')


class PathEnsembleTest(unittest.TestCase):
    """Run a test for the PathEnsemble class."""

    def test_logic(self):
        """Test eq and ne."""
        path_ensemble1 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        path_ensemble2 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        path_ensemble2.rgen = path_ensemble1.rgen
        self.assertFalse(path_ensemble1 == 'Nonno')
        self.assertTrue(path_ensemble2 == path_ensemble1)

        path_ensemble2 = PathEnsemble(0, [-2, 0, 1], maxpath=10, exe_dir=None)
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)

        path_ensemble2 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        path_ensemble2.ciccio = None
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)

        path_ensemble2 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        path_ensemble2.directory = {'accepted': 'GranaPadano'}
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)

        path_ensemble2 = PathEnsemble(0, [-1, 0, 1], maxpath=16, exe_dir=None)
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)

        path_ensemble2 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        path_ensemble2.last_path = Path()
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)

        path_ensemble1 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        path_ensemble2 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        path_ensemble1.rgen = create_random_generator({'rgen': 'rgen',
                                                       'seed': 123})
        path_ensemble2.rgen = create_random_generator({'rgen': 'rgen',
                                                       'seed': 123})
        path_ensemble2.rgen.rand()
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)
        path_ensemble1.rgen = create_random_generator({'rgen': 'rgen',
                                                       'seed': 132})
        path_ensemble2.rgen = create_random_generator({'rgen': 'rgen',
                                                       'seed': 123})
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)

        path_ensemble2 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        path_ensemble1.rgen = create_random_generator({'rgen': 'mock'})
        path_ensemble2.rgen = create_random_generator({'rgen': 'rgen'})
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)

        path_ensemble1 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        path_ensemble2 = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)

        path_ensemble1.last_path = Path()
        path_ensemble2.last_path = [1, 2]
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)
        path_ensemble1.last_path = PATHTEST0
        path_ensemble2.last_path = PATHTEST1
        with patch('sys.stdout', new=StringIO()):
            self.assertTrue(path_ensemble2 != path_ensemble1)

    def test_init(self):
        """Test initiation."""
        path_ensemble = PathEnsemble(0, [-1, 0, 1], maxpath=10, exe_dir=None)
        self.assertEqual(path_ensemble.start_condition, 'R')
        self.assertEqual(path_ensemble.ensemble_name_simple, '000')
        path_ensemble = PathEnsemble(11, [-1, 0, 1], maxpath=10, exe_dir=None)
        self.assertEqual(path_ensemble.start_condition, 'L')
        self.assertEqual(path_ensemble.ensemble_name_simple, '011')

    def test_directories(self):
        """Test if we get back a directory."""
        path_ensemble1 = PathEnsemble(0, [-1, 0, 1])
        path_ensemble2 = PathEnsemble(0, [-1, 0, 1], exe_dir=DIR_NAME)
        j = 0
        for i in path_ensemble1.directories():
            self.assertTrue(i is None)
            j += 1
        self.assertEqual(j, 4)
        j = 0
        dirs = [i for i in path_ensemble2.directories()]
        self.assertTrue(os.path.join(DIR_NAME, '000') in dirs)
        j += 1
        for i in ('accepted', 'generate', 'traj'):
            testname = os.path.join(DIR_NAME, '000', i)
            self.assertTrue(testname in dirs)
            j += 1
        self.assertEqual(j, 4)

    def test_update_dir(self):
        """Test that we can update directories."""
        path_ensemble = PathEnsemble(1, [-1, 0, 1], exe_dir='test1')
        old = [i for i in path_ensemble.directories()]
        path_ensemble.update_directories(os.path.join('test1', '002'))
        new = [i for i in path_ensemble.directories()]
        for diri, dirj in zip(old, new):
            for i, j in zip(diri, dirj):
                if i != j:
                    self.assertEqual(i, '1')
                    self.assertEqual(j, '2')

    def test_add_path(self):
        """Test adding of paths and reset."""
        path_ensemble = PathEnsemble(1, [-1, 0, 1], maxpath=10)
        # Add a path
        path = make_fake_path(length=10)
        path_ensemble.add_path_data(path, 'ACC', cycle=0)
        self.assertEqual(path_ensemble.nstats['npath'], 1)
        self.assertEqual(path_ensemble.nstats['ACC'], 1)
        self.assertTrue(path is path_ensemble.last_path)
        # Add empty path:
        path_ensemble.add_path_data(None, 'KOB', cycle=1)
        self.assertEqual(path_ensemble.nstats['npath'], 2)
        self.assertEqual(path_ensemble.nstats['KOB'], 1)
        self.assertTrue(path is path_ensemble.last_path)
        # Add for a shooting move:
        path = make_fake_path(length=3)
        path.generated = ('sh', 1, 2, 3)
        path_ensemble.add_path_data(path, 'ACC', cycle=2)
        for _ in range(7):
            path_ensemble.add_path_data(path, 'ACC')
        self.assertEqual(len(path_ensemble.paths), 10)
        path_ensemble.add_path_data(path, 'ACC')
        self.assertEqual(len(path_ensemble.paths), 1)
        path_ensemble.reset_data()
        self.assertEqual(len(path_ensemble.paths), 0)
        for _, val in path_ensemble.nstats.items():
            self.assertEqual(val, 0)

    def test_looping(self):
        """Test adding of paths and looping."""
        path_ensemble = PathEnsemble(1, [-1, 0, 1], maxpath=20)
        correct = []  # For storing the correct lengths.
        for i in range(5):
            correct.append(10 + i)
            path = make_fake_path(length=10+i)
            path_ensemble.add_path_data(path, 'ACC')
        path_ensemble.add_path_data(None, 'KOB')
        correct.append(14)
        path_ensemble.add_path_data(None, 'KOB')
        correct.append(14)
        for i in range(3):
            correct.append(10 + i + 5)
            path = make_fake_path(length=10+i+5)
            path_ensemble.add_path_data(path, 'ACC')
        for _ in range(5):
            path_ensemble.add_path_data(None, 'KOB')
            correct.append(10 + 2 + 5)  # "+2" is from the previous for loop.
        for i, path in enumerate(path_ensemble.get_paths()):
            if i in (5, 6, 10, 11, 12, 13, 14):
                self.assertEqual(path['status'], 'KOB')
            else:
                self.assertEqual(path['status'], 'ACC')
        for i, path in enumerate(path_ensemble.get_accepted()):
            self.assertEqual(path['length'], correct[i])
        self.assertAlmostEqual(path_ensemble.get_acceptance_rate(), 8./15.)

    def test_restart_info1(self):
        """Test that we can make restart info."""
        path_ensemble = PathEnsemble(1, [-1, 0, 1], maxpath=20)
        for i in range(5):
            path = make_fake_path(length=10+i)
            path_ensemble.add_path_data(path, 'ACC')
        path_ensemble.add_path_data(None, 'KOB')
        path_ensemble.add_path_data(None, 'KOB')
        info = path_ensemble.restart_info()
        path_ensemble2 = PathEnsemble(10, [0, 0, 0], maxpath=1)
        path_ensemble2.load_restart_info(info)
        # Note we do not force interfaces when loading restart,
        # here, just check that nstats were correctly loaded:
        for key, val in path_ensemble.nstats.items():
            self.assertEqual(val, path_ensemble2.nstats[key])

    def test_restart_info2(self):
        """Test that we can make restart info."""
        path_ensemble = PathEnsemble(1, [-1, 0, 1])
        for i in range(5):
            path = make_fake_path(length=10+i)
            path_ensemble.add_path_data(path, 'ACC')
        path_ensemble.add_path_data(None, 'KOB')
        path_ensemble.add_path_data(None, 'KOB')
        info = path_ensemble.restart_info()
        path_ensemble2 = PathEnsemble(10, [0, 0, 0])
        path_ensemble2.load_restart_info(info)

        # Note we do not force interfaces when loading restart
        # information, here, just check that nstats were
        # correctly loaded:
        for key, val in path_ensemble.nstats.items():
            self.assertEqual(val, path_ensemble2.nstats[key])


class PathEnsembleExtTest(unittest.TestCase):
    """Run a test for the PathEnsembleExt class."""

    def setUp(self):
        """Just make sure we create needed directories."""
        create_dirs()

    def tearDown(self):
        """Remove the created directories."""
        remove_dirs()

    def test_init(self):
        """Test initiation."""
        ens = PathEnsembleExt(1, [-1., 0., 1.], exe_dir=DIR_NAME)
        correct_dir = [
            os.path.join(DIR_NAME, '001'),
            os.path.join(DIR_NAME, '001', 'accepted'),
            os.path.join(DIR_NAME, '001', 'generate'),
            os.path.join(DIR_NAME, '001', 'traj'),
        ]
        for dirname, correct in zip(ens.directories(), correct_dir):
            self.assertEqual(dirname, correct)

    def test_update_dir(self):
        """Test that we can update directories."""
        path_ensemble = PathEnsembleExt(1, [-1, 0, 1], exe_dir='test1')
        old = [i for i in path_ensemble.directories()]
        path_ensemble.update_directories(os.path.join('test1', '002'))
        new = [i for i in path_ensemble.directories()]
        for diri, dirj in zip(old, new):
            for i, j in zip(diri, dirj):
                if i != j:
                    self.assertEqual(i, '1')
                    self.assertEqual(j, '2')

    def test_move_path(self):
        """Test that we can move paths."""
        with tempfile.TemporaryDirectory() as tempdir:
            ens = PathEnsembleExt(1, [-1., 0., 1.], exe_dir=tempdir)
            for i in ens.directories():
                make_dirs(i)
            for name in ens.list_superfluous():
                os.remove(name)
            path = make_fake_path_files(tempdir)
            # Add a file so that we will have to overwrite it:
            target = os.path.join(tempdir, '001', 'accepted', 'fake_path_2')
            with open(target, 'w', encoding='utf-8') as output:
                output.write('Blekkulf, er du der?')
            # And add a file we don't need:
            target = os.path.join(tempdir, '001', 'accepted', 'extra-file')
            with open(target, 'w', encoding='utf-8') as output:
                output.write('Takpapp, veggpapp, gulvpapp, tapet')
            file_paths = []
            for i in path.phasepoints:
                file_paths.append(i.particles.get_pos()[0])
            # Add the file as accepted. This will move it:
            ens.add_path_data(path, 'ACC', cycle=0)
            file_paths2 = []
            for i in path.phasepoints:
                file_paths2.append(i.particles.get_pos()[0])
            for i, j in zip(file_paths, file_paths2):
                # Check that files were moved:
                self.assertNotEqual(i, j)
                self.assertFalse(os.path.isfile(i))
                self.assertTrue(os.path.isfile(j))
                # Check that we did not alter the base name:
                self.assertEqual(os.path.basename(i), os.path.basename(j))
                # Check that files were moved into the accepted folder:
                target = os.path.join(
                    ens.directory['accepted'], os.path.basename(j)
                )
                self.assertEqual(j, target)
            # Force move path, when source and target are the same:
            ens.add_path_data(path, 'ACC', cycle=0)

    def test_move_to_generate(self):
        """Test that we can move a path to the generated folder."""
        with tempfile.TemporaryDirectory() as tempdir:
            ens = PathEnsembleExt(1, [-1., 0., 1.], exe_dir=tempdir)
            for i in ens.directories():
                make_dirs(i)
            path = make_fake_path_files(tempdir)
            # Check that generated files exists:
            for i in path.phasepoints:
                self.assertTrue(os.path.isfile(i.particles.get_pos()[0]))
            # Add to path ensemble (and move to the accepted folder):
            ens.add_path_data(path, 'ACC', cycle=0)
            for i in path.phasepoints:
                target = os.path.join(
                    tempdir, ens.directory['accepted'],
                    os.path.basename(i.particles.get_pos()[0])
                )
                # After the move, these names should be equal:
                self.assertEqual(target, i.particles.get_pos()[0])
            ens.move_path_to_generate(path, prefix='gen_')
            # Check that files were moved to the generated folder:
            for i in path.phasepoints:
                self.assertTrue(os.path.isfile(i.particles.get_pos()[0]))
                target = os.path.join(
                    tempdir, ens.directory['generate'],
                    os.path.basename(i.particles.get_pos()[0])
                )
                self.assertEqual(target, i.particles.get_pos()[0])

    def test_copy_path(self):
        """Test that we can copy a path."""
        with tempfile.TemporaryDirectory() as tempdir:
            ens = PathEnsembleExt(1, [-1., 0., 1.], exe_dir=tempdir)
            for i in ens.directories():
                make_dirs(i)
            path = make_fake_path_files(tempdir)
            ens.add_path_data(path, 'ACC', cycle=0)
            target_dir = ens.directory['path_ensemble']
            path_copy = ens._copy_path(  # pylint: disable=protected-access
                path,
                target_dir,
                prefix='copy_'
            )
            # Check that files were copied:
            for i in path_copy.phasepoints:
                self.assertTrue(os.path.isfile(i.particles.get_pos()[0]))
            # Check that original files remains:
            for i in path.phasepoints:
                self.assertTrue(os.path.isfile(i.particles.get_pos()[0]))
            # Check that original and copy path are different:
            for i, j in zip(path.phasepoints, path_copy.phasepoints):
                self.assertNotEqual(i.particles.get_pos()[0],
                                    j.particles.get_pos()[0])
            # Test that delete then copy function of _copy_path:
            rewrite_file = tempdir + '/001/copy_fake_path_1'
            # Empties file of text:
            open(rewrite_file, 'w', encoding='utf-8').close()
            self.assertEqual(os.path.getsize(rewrite_file), 0)
            ens._copy_path(  # pylint: disable=protected-access
                path,
                target_dir,
                prefix='copy_'
            )
            self.assertNotEqual(os.path.getsize(rewrite_file), 0)
            # Tests copy_path_to_generate():
            ens.copy_path_to_generate(path)
            for i in path.phasepoints:
                i_path = i.particles.get_pos()[0].replace('accepted',
                                                          'generate')
                self.assertTrue(os.path.isfile(i_path))
            # Check that original files remains:
            for i in path.phasepoints:
                self.assertTrue(os.path.isfile(i.particles.get_pos()[0]))

    def test_restart(self):
        """Test that we can write/read restart info."""
        with tempfile.TemporaryDirectory() as tempdir:
            ens = PathEnsembleExt(1, [-1., 0., 1.], exe_dir=tempdir)
            for i in ens.directories():
                make_dirs(i)
            path = make_fake_path_files(tempdir)
            ens.add_path_data(path, 'ACC', cycle=0)
            info = ens.restart_info()
            ens2 = PathEnsembleExt(2, [-1., 0.5, 1.], exe_dir=tempdir)
            for i in ens2.directories():
                make_dirs(i)
            # Note that this will NOT copy any paths, just set some path
            # names. We just check that we get a warning about this:
            with turn_on_logging():
                with self.assertLogs('pyretis.core.pathensemble',
                                     level='CRITICAL'):
                    with self.assertLogs('pyretis.core.random_gen',
                                         level='CRITICAL'):
                        ens2.load_restart_info(info)

    def test_clear_generate(self):
        """Test that we can clear the ens/generate folder."""
        with tempfile.TemporaryDirectory() as tempdir:
            ens = PathEnsembleExt(1, [-1., 0., 1.], exe_dir=tempdir)
            for i in ens.directories():  # creating 001
                make_dirs(i)  # creating traj, generate
            # make two paths in 001/generate
            make_fake_path_files(tempdir+"/001/generate")
            # first assert that the generate folder is not empty
            self.assertTrue(os.listdir(tempdir+"/001/generate"))
            # clear generate folder
            ens.clear_generate()
            # assert that the generate folder is now empty.
            self.assertFalse(os.listdir(tempdir+"/001/generate"))

    def test_print_pe(self):
        """Test that printing path ensemble information does not crash."""
        with tempfile.TemporaryDirectory() as tempdir:
            ens = PathEnsembleExt(1, [-1., 0., 1.], exe_dir=tempdir)
            for i in ens.directories():
                make_dirs(i)
            path = make_fake_path_files(tempdir)
            ens.add_path_data(path, 'ACC', cycle=0)
            txt = str(ens)
            ref = 'Path ensemble: [0^+]\n\tInterfaces: (-1.0, 0.0, 1.0)\n\t'
            ref += 'Number of paths stored: 1\n\tNumber of accepted paths: 1'
            ref += '\n\tRatio accepted/total paths: 1.0'
            self.assertEqual(txt, ref)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test Monte Carlo Methods"""
import logging
import unittest
import numpy as np
from pyretis.core.system import System
from pyretis.core.particles import Particles
from pyretis.forcefield.potentials import DoubleWell
from pyretis.forcefield import ForceField
from pyretis.core.montecarlo import (
    metropolis_accept_reject,
    accept_reject_momenta,
    max_displace_step,
)
from pyretis.core.random_gen import MockRandomGenerator
logging.disable(logging.CRITICAL)


class MonteCarloTest(unittest.TestCase):
    """Run test for the Monte Carlo module"""

    def test_metropolis(self):
        """Test that we can draw random numbers in [0, 1)"""
        rgen = MockRandomGenerator(seed=0)
        system = System(units='reduced', temperature=1)
        self.assertTrue(metropolis_accept_reject(rgen, system, -1.0))
        acc = metropolis_accept_reject(rgen, system, 1.0)
        self.assertFalse(acc)

    def test_momenta(self):
        """Test the accept/reject for momenta."""
        rgen = MockRandomGenerator(seed=0)
        system = System(units='reduced', temperature=1)
        acc = accept_reject_momenta(rgen, system, 10.0, aimless=True)
        self.assertTrue(acc)
        acc = accept_reject_momenta(rgen, system, 10.0, aimless=False)
        self.assertFalse(acc)

    def test_max_displace(self):
        """Test the max-displace."""
        rgen = MockRandomGenerator(seed=0)
        system = System(units='reduced', temperature=1)
        particles = Particles(dim=1)
        particles.add_particle(np.zeros(1), np.zeros(1), np.zeros(1))
        particles.add_particle(np.zeros(1), np.zeros(1), np.zeros(1))
        system.particles = particles
        potential_dw = DoubleWell(a=1, b=1, c=0.02)
        system.forcefield = ForceField(
            'Test',
            potential=[potential_dw],
            params=[{'a': 1.0, 'b': 1.0, 'c': 0.02}]
        )
        system.potential_and_force()
        acc = max_displace_step(rgen, system, maxdx=10., idx=None)
        self.assertFalse(acc[-1])
        acc = max_displace_step(rgen, system, maxdx=1.25, idx=0)
        acc = max_displace_step(rgen, system, maxdx=1.25, idx=0)
        self.assertTrue(acc[-1])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""PyRETIS style for use with pygments.

This style is based on `Xcode` from the Pygments.
For the original style:

    :pyretis_copyright: Copyright 2006-2029 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.

See: http://pygments.org/
"""

from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String, Error, \
     Number, Operator, Literal, Generic, Punctuation


def setup(app):
    """Register the script."""
    pass


class PyretisStyle(Style):
    """Style similar to the Xcode default colouring theme."""

    default_style = '#262626'

    styles = {
        Comment:                '#177500',
        Comment.Preproc:        '#633820',
        String:                 '#C41A16',
        String.Char:            '#2300CE',
        Operator:               '#262626',
        Keyword:                '#A90D91',
        Name:                   '#262626',
        Name.Attribute:         '#836C28',
        Name.Class:             '#3F6E75',
        Name.Function:          '#262626',
        Name.Builtin:           '#A90D91',
        Name.Builtin.Pseudo:    '#5B269A',
        Name.Variable:          '#262626',
        Name.Tag:               '#262626',
        Name.Decorator:         '#262626',
        Name.Label:             '#262626',
        Punctuation:            '#262626',
        Literal:                '#1C01CE',
        Number:                 '#1C01CE',
        Error:                  '#262626',
        Generic:                "#262626",
        Generic.Deleted:        "#A00000",
        Generic.Emph:           "italic",
        Generic.Error:          "#FF0000",
        Generic.Heading:        "bold #000080",
        Generic.Inserted:       "#00A000",
        Generic.Output:         "#262626",
        Generic.Prompt:         "bold #000080",
        Generic.Strong:         "bold",
        Generic.Subheading:     "bold #800080",
        Generic.Traceback:      "#0044DD",
    }



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the functionality of units from pyretis.core.

We also do some test to check the correctness of values.
"""
from copy import deepcopy
import logging
import os
import tempfile
import unittest
from pyretis.core.units import (
    create_conversion_factors,
    generate_system_conversions,
    generate_conversion_factors,
    read_conversions,
    CONVERT,
    UNITS,
    write_conversions,
    units_from_settings,
    convert_bases,
    bfs_convert,
)
from .help import turn_on_logging
logging.disable(logging.CRITICAL)


class UnitsTest(unittest.TestCase):
    """Run the tests of methods from pyretis.core.units."""

    def test_create_lennard_jones_units(self):
        """Test that we create correct Lennard-Jones units."""
        create_conversion_factors('lj', length=(3.405, 'A'),
                                  energy=(119.8, 'kB'),
                                  mass=(39.948, 'g/mol'),
                                  charge='e')
        self.assertAlmostEqual(CONVERT['length']['bohr', 'nm'],
                               0.052917721067000, 12)
        self.assertAlmostEqual(CONVERT['length']['lj', 'nm'],
                               0.34050, 4)
        self.assertAlmostEqual(CONVERT['length']['bohr', 'lj'],
                               0.155411809301, 12)
        self.assertAlmostEqual(CONVERT['time']['lj', 'ps'],
                               2.156349772323142, 12)
        self.assertAlmostEqual(CONVERT['mass']['lj', 'kg'],
                               6.633521358698435e-26, 12)
        self.assertAlmostEqual(CONVERT['force']['lj', 'N'],
                               4.857612120293684e-12, 12)
        self.assertAlmostEqual(CONVERT['energy']['lj', 'J'],
                               1.654016926960000e-21, 12)
        self.assertAlmostEqual(CONVERT['velocity']['lj', 'nm/ps'],
                               1.579057369867981e-01, 12)
        self.assertAlmostEqual(CONVERT['pressure']['lj', 'bar'],
                               4.189754740302599e+02, 12)
        self.assertAlmostEqual(CONVERT['temperature']['lj', 'K'],
                               1.198000000000000e+02, 12)
        self.assertAlmostEqual(CONVERT['charge']['lj', 'e'],
                               4.940801883873017e-02, 12)

    def test_create_cgs_units(self):
        """Test that we create correct cgs units."""
        create_conversion_factors('cgs', length=(0.01, 'm'),
                                  energy=(1.0e-7, 'J'),
                                  mass=(1.0, 'g'), charge='e')
        self.assertAlmostEqual(CONVERT['force']['cgs', 'dyn'],
                               1.0, 12)
        self.assertAlmostEqual(CONVERT['force']['cgs', 'N'],
                               1.0e-5, 12)
        self.assertAlmostEqual(CONVERT['time']['cgs', 's'],
                               1.0, 12)
        self.assertAlmostEqual(CONVERT['length']['cgs', 'm'],
                               1.0e-2, 12)
        self.assertAlmostEqual(CONVERT['mass']['cgs', 'kg'],
                               1.0e-3, 12)
        self.assertAlmostEqual(CONVERT['velocity']['cgs', 'm/s'],
                               1.0e-2, 12)
        self.assertAlmostEqual(CONVERT['energy']['cgs', 'J'],
                               1.0e-7, 12)
        self.assertAlmostEqual(CONVERT['pressure']['cgs', 'Pa'],
                               1.0e-1, 12)
        self.assertAlmostEqual(CONVERT['temperature']['cgs', 'K'],
                               1.0, 12)
        self.assertAlmostEqual(CONVERT['charge']['cgs', 'C'],
                               3.335640951981520e-10, 12)

    def test_have_common_unit_systems(self):
        """Test that we can generate all conversions."""
        systems = ('lj', 'real', 'metal', 'au',
                   'electron', 'si', 'gromacs')
        for sys in systems:
            create_conversion_factors(sys)
        all_pairs = []
        for sys1 in systems:
            for sys2 in systems:
                if sys1 != sys2:
                    generate_system_conversions(sys1, sys2)
                    all_pairs.append((sys1, sys2))
        for key in CONVERT:
            msg = ['Could not find conversion "{}" -> "{}"',
                   f'for dimension "{key}"']
            for pair in all_pairs:
                msg[0] = msg[0].format(*pair)
                msgtxt = ' '.join(msg)
                self.assertIn(pair, CONVERT[key], msg=msgtxt)

    def test_creation_of_units(self):
        """Test that creation of units works and fails as expected."""
        with self.assertRaises(ValueError):
            create_conversion_factors('test', length=None, energy=None,
                                      mass=None, charge=None)
        with self.assertRaises(ValueError):
            create_conversion_factors('test', length=(1.0, 'm'),
                                      energy=(1.0, 'J'), mass=(1.0, 'kg'),
                                      charge=None)
        with self.assertRaises(ValueError):
            create_conversion_factors('test', length=(1.0, 'm'),
                                      energy=(1.0, 'J'), mass=(1.0, 'kg'),
                                      charge='a non-existing unit')
        with self.assertRaises(LookupError):
            create_conversion_factors('test', length=(1.0, 'strange_unit'),
                                      energy=(1.0, 'J'), mass=(1.0, 'kg'),
                                      charge='e')
        with self.assertRaises(TypeError):
            create_conversion_factors('test', length=1.0,
                                      energy=(1.0, 'J'), mass=(1.0, 'kg'),
                                      charge='e')
        # the next one should be successful
        create_conversion_factors('test', length=(1.0, 'm'),
                                  energy=(1.0, 'J'), mass=(1.0, 'kg'),
                                  charge='e')
        generate_system_conversions('test', 'real')
        # check if we indeed created all conversions
        for key in CONVERT:
            dimtxt = f'for dimension "{key}"'
            for unit in UNITS[key]:
                pair = ('test', unit)
                msg = 'Could not find conversion "{}" -> "{}"'.format(*pair)
                msgtxt = ' '.join([msg, dimtxt])
                self.assertIn(pair, CONVERT[key], msgtxt)
                pair = (unit, 'test')
                msg = 'Could not find conversion "{}" -> "{}"'.format(*pair)
                msgtxt = ' '.join([msg, dimtxt])
                self.assertIn(pair, CONVERT[key], msgtxt)

        with self.assertRaises(ValueError):
            create_conversion_factors('test', length=(1.0, 'm'),
                                      energy=(1.0, 'J'),
                                      mass=(1.0, 'kg'), charge=(100, 'e'))

    def test_read_from_file(self):
        """Test that we can read units from a input file."""
        dirname = os.path.dirname(os.path.abspath(__file__))
        filename = os.path.join(dirname, 'units_input.txt')
        conv = read_conversions(filename=filename,
                                select_units='test_system')
        self.assertAlmostEqual(conv['charge']['C', 'test_system'],
                               1.23450, 12)
        self.assertAlmostEqual(conv['energy']['J', 'test_system'],
                               6.7890, 12)
        self.assertAlmostEqual(conv['force']['N', 'test_system'],
                               4.2424242e1, 12)
        self.assertAlmostEqual(conv['length']['m', 'test_system'],
                               1.111112e1, 12)
        self.assertAlmostEqual(conv['mass']['kg', 'test_system'],
                               1.234e-4, 12)
        self.assertAlmostEqual(conv['pressure']['Pa', 'test_system'],
                               1e-1, 12)
        self.assertAlmostEqual(conv['temperature']['K', 'test_system'],
                               1.0, 12)
        self.assertAlmostEqual(conv['time']['s', 'test_system'],
                               0.5, 12)
        self.assertAlmostEqual(conv['velocity']['m/s', 'test_system'],
                               2.0, 12)

    def test_write_conversions(self):
        """Test that we can write out the conversions."""
        with tempfile.NamedTemporaryFile() as temp:
            write_conversions(filename=temp.name)
            temp.flush()
            conv = read_conversions(filename=temp.name)
            for key, val in conv.items():
                self.assertTrue(key in CONVERT)
                self.assertAlmostEqual(val, CONVERT[key])

    def test_convert_bases(self):
        """Test convert_bases function"""
        copy_convert = deepcopy(CONVERT)
        CONVERT['mass'] = {('kg', 'kg'): 1.0, ('g', 'kg'): 0.001,
                           ('g', 'g'): 1.0}
        with turn_on_logging():
            with self.assertLogs('pyretis.core.units', level='WARNING'):
                convert_bases('mass')
        self.assertAlmostEqual(CONVERT['mass'][('kg', 'g')], 1000.0)
        CONVERT.clear()
        CONVERT.update(copy_convert)

    def test_bfs_convert(self):
        """Test bfs_convert function"""
        convertdim = {('g', 'g'): 1.0, ('kg', 'g'): 1000.0, ('kg', 'kg'): 1.0}
        self.assertEqual(bfs_convert(convertdim, 'g', 'kg'),
                         (('g', 'kg'), None, None))
        self.assertEqual(bfs_convert(convertdim, 'kg', 'kg'),
                         (('kg', 'kg'), 1.0, None))
        self.assertEqual(bfs_convert(convertdim, 'kg', 'g'),
                         (('kg', 'g'), 1000.0, None))

    def test_generate_conversion(self):
        """Test the generator of conversion factors."""
        generate_conversion_factors('myunit', (1.0, 'm'), (1.0, 'J'),
                                    (1.0, 'kg'), charge='e')
        self.assertAlmostEqual(CONVERT['mass']['myunit', 'g'], 1000.)
        generate_conversion_factors('myunit2', (1.0, 'm'), (1.0, 'J'),
                                    (1.0, 'kg'), charge='C')
        # Test with missing kB
        with self.assertRaises(ValueError):
            generate_conversion_factors('myunit3', (1.0, 'm'), (1.0, 'JJ'),
                                        (1.0, 'kg'), charge='e')
        # Test with non-unit "1.0" values
        generate_conversion_factors('myunit4', (2.0, 'm'), (0.5, 'J'),
                                    (3.0, 'g'))
        self.assertAlmostEqual(CONVERT['length']['myunit4', 'm'], 2.0)
        self.assertAlmostEqual(CONVERT['length']['m', 'myunit4'], 0.5)
        self.assertAlmostEqual(CONVERT['energy']['myunit4', 'J'], 0.5)
        self.assertAlmostEqual(CONVERT['energy']['J', 'myunit4'], 2.0)
        self.assertAlmostEqual(CONVERT['mass']['myunit4', 'g'], 3.0)
        self.assertAlmostEqual(CONVERT['mass']['g', 'myunit4'], 0.33333333)
        self.assertAlmostEqual(CONVERT['mass']['myunit4', 'kg'], 0.003)
        # Test with 'unit' label part of  UNITS[dim]. To check coverage.
        copy_convert = deepcopy(CONVERT)
        self.assertEqual(CONVERT, copy_convert)
        generate_conversion_factors('bohr', (1., 'm'), (1., 'J'), (1., 'g'))
        self.assertNotEqual(CONVERT, copy_convert)
        self.assertAlmostEqual(CONVERT['length']['bohr', 'bohr'], 1.)
        # Test some expected but wrong assignment
        self.assertAlmostEqual(CONVERT['length']['bohr', 'm'], 1.)
        self.assertAlmostEqual(CONVERT['mass']['bohr', 'g'], 1.)
        # Restore CONVERT
        CONVERT.clear()
        CONVERT.update(copy_convert)
        self.assertEqual(CONVERT, copy_convert)
        with self.assertRaises(Exception) as context:
            generate_conversion_factors('au', (1., 'm'), (1., 'J'), (1., 'xx'))
        self.assertTrue("Missing" in str(context.exception))

    def test_units_from_settings(self):
        """Test that we can create units from settings."""
        # Using a predefined unit system
        settings = {'system': {'units': 'real'}}
        msg = units_from_settings(settings)
        self.assertEqual(msg, 'Created units: "real".')
        # Using a custom system:
        settings = {'system': {'units': 'my_new_system1'}}
        settings['unit-system'] = {
            'name': 'my_new_system1',
            'length': (1.0, 'bohr'),
            'mass': (1.0, 'g'),
            'energy': (1.0, 'J'),
            'charge': 'e'
        }
        msg = units_from_settings(settings)
        self.assertEqual(msg, 'Created unit system: "my_new_system1".')
        # Test for errors:
        with self.assertRaises(ValueError):
            settings = {'system': {'units': 'my_new_system2'}}
            settings['unit-system'] = {
                'length': (1.0, 'bohr'),
                'mass': (1.0, 'g'),
                'energy': (1.0, 'J'),
                'charge': 'e'
            }
            units_from_settings(settings)
        # Inconsistent name:
        with self.assertRaises(ValueError):
            settings = {'system': {'units': 'my_new_system2'}}
            settings['unit-system'] = {
                'name': 'my_new_system3',
                'length': (1.0, 'bohr'),
                'mass': (1.0, 'g'),
                'energy': (1.0, 'J'),
                'charge': 'e'
            }
            units_from_settings(settings)
        # Missing one of length, mass, energy, charge
        for i in ('length', 'mass', 'energy', 'charge'):
            settings = {'system': {'units': 'my_new_system2'}}
            settings['unit-system'] = {
                'name': 'my_new_system2',
                'length': (1.0, 'bohr'),
                'mass': (1.0, 'g'),
                'energy': (1.0, 'J'),
                'charge': 'e'
            }
            del settings['unit-system'][i]
            with self.assertRaises(ValueError):
                units_from_settings(settings)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Just plot some results using bokeh."""
import numpy as np
from bokeh.plotting import figure, output_file, show
from bokeh.models import Range1d, LinearAxis
from bokeh.palettes import all_palettes
from bokeh.layouts import gridplot


def make_plot():
    """Create plot using bokeh."""
    colors = all_palettes['Set1'][4]

    raw_data = np.loadtxt('thermo.txt')
    data = {}
    for i, key in enumerate(('step', 'temp', 'pot', 'kin', 'tot', 'press')):
        data[key] = raw_data[:, i]

    output_file('thermo_plot.html')
    tot0 = data['tot'][0]
    plt1 = figure(x_range=Range1d(data['step'][0], data['step'][-1]*1.05),
                  width=400, height=400, toolbar_location='above')
    plt1.line(data['step'], data['pot']-tot0, line_width=3,
              line_color=colors[0], alpha=0.8, legend='Potential')
    plt1.line(data['step'], data['kin']-tot0, line_width=3,
              line_color=colors[1], alpha=0.8, legend='Kinetic')
    plt1.line(data['step'], data['tot']-tot0, line_width=3,
              line_color=colors[2], alpha=0.8, legend='Total')

    plt1.xaxis.axis_label = 'Step'
    plt1.yaxis.axis_label = 'Energy per particle'
    plt1.legend.location = 'center_left'
    plt1.legend.label_text_font_size = '12pt'
    plt1.background_fill_color = '#E8DDCB'
    plt1.background_fill_alpha = 0.5
    plt1.xaxis.axis_label_text_font_size = '12pt'
    plt1.yaxis.axis_label_text_font_size = '12pt'
    plt1.xaxis.major_label_text_font_size = '12pt'
    plt1.yaxis.major_label_text_font_size = '12pt'

    plt2 = figure(x_range=Range1d(data['step'][0], data['step'][-1]),
                  width=400, height=400, toolbar_location='above')
    plt2.line(data['step'], data['press'], line_width=3,
              line_color=colors[1], alpha=0.8, legend='Pressure')
    plt2.xaxis.axis_label = 'Step'
    plt2.yaxis.axis_label = 'Pressure'
    plt2.legend.location = 'center_left'
    plt2.background_fill_color = '#E8DDCB'
    plt2.background_fill_alpha = 0.5

    plt2.extra_y_ranges = {'temp': Range1d(start=data['temp'].min(),
                                           end=data['temp'].max()*1.05)}
    plt2.line(data['step'], data['temp'], line_width=3,
              line_color=colors[2], alpha=0.8, legend='Temperature',
              y_range_name='temp')
    plt2.add_layout(LinearAxis(y_range_name='temp', axis_label='Temperature'),
                    'right')
    plt2.xaxis.axis_label_text_font_size = '12pt'
    plt2.yaxis.axis_label_text_font_size = '12pt'
    plt2.legend.label_text_font_size = '12pt'

    plt2.xaxis.major_label_text_font_size = '12pt'
    plt2.yaxis.major_label_text_font_size = '12pt'

    subplot = gridplot(
        (plt1, plt2),
        ncols=2,
        plot_width=400,
        plot_height=400,
    )
    show(subplot)


if __name__ == '__main__':
    make_plot()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Collection of statistical methods for PyVisA.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
correlation_matrix (:py:func:`.correlation_matrix`)
    Method for displaying the correlations from the simulation variables.

gaussian_mixture (:py:func:`.gaussian_mixture`)
    Generic method for performing gaussian mixture clustering
    on simulation data.

hierarchical (:py:func:`.hierarchical`)
    Generic method for performing hierarchical clustering on simulation data.

k_means (:py:func:`.k_means`)
    Generic method for performing K-means clustering on simulation data.

pyvis_pca (:py:func:`.pyvisa_pca`)
    Generic method for performing pca on simulation data.

spectral (:py:func:`.spectral`)
    Generic method for performing spectral clustering on simulation data.

"""
import graphviz
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from sklearn.cluster import AgglomerativeClustering, SpectralClustering, \
    MiniBatchKMeans
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestClassifier
from sklearn.mixture import GaussianMixture
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier, export_graphviz


def _plot_pca_results(n_pca, principal_df, loadings, pca_info, cmap):
    """Plot the results of a PCA analysis.

    Parameters
    ----------
    n_pca : integer
        Number of principal components.
    principal_df : object like pandas.DataFrame
        Principal components data.
    loadings : object like pandas.DataFrame
        PCA loadings.
    pca_info : dict
        Dict with keys 'model' (fitted PCA), 'cols' (column labels),
        'features' (feature labels).
    cmap : string
        Matplotlib colormap.

    """
    pca_model = pca_info['model']
    cols = pca_info['cols']
    features = pca_info['features']
    # Plot PC1 vs PC2
    if n_pca > 1:
        fig_1 = plt.figure()
        axis_1 = fig_1.add_subplot(111)
        axis_1.scatter(principal_df['PC1'], principal_df['PC2'],
                       edgecolors=None, alpha=0.5,
                       cmap=cmap)
        axis_1.set_xlabel('PC1')
        axis_1.set_ylabel('PC2')

    # Plot the loadings
    fig_2 = plt.figure()
    axis_2 = fig_2.add_subplot(111)
    fig_2.colorbar(axis_2.matshow(loadings))
    axis_2.set_xticks(np.arange(len(cols)))
    axis_2.set_yticks(np.arange(len(features)))
    axis_2.set_xticklabels(cols)
    axis_2.set_yticklabels(features)

    # Plot Cumulative explained variance
    fig_3 = plt.figure()
    axis_3 = fig_3.add_subplot(111)
    axis_3.plot(np.cumsum(pca_model.explained_variance_ratio_))
    axis_3.set_xlabel('Principal components')
    axis_3.set_ylabel('Cumulative explained variance')
    plt.show()


def pyvisa_pca(n_pca, settings, data, cmap):
    """Perform PCA on ensemble data.

    This function performs PCA on data from the selected ensembles
    and stores the data to a hdf-file for further analysis, plots the
    first two principal components against each other,
    the cumulative explained variance and the loadings.
    The hdf5 file contains the keywords:

    * `sim_data`: pandas Dataframe containing the simulation data.
    * `PC`: Dataframe containing the data from the principal components.
    * `loadings`: pandas Dataframe containing the loadings.
    * `explained`: pandas Series containing the simulation data.
    * `correlation_matrix`: Dataframe containing the correlation matrix.


    Parameters
    ----------
    n_pca : integer
        Number of clusters.
    settings : dict
        Settings from GUI.
    data : object like pandas.DataFrame
        Simulation data from chosen ensembles.
    cmap : string
        Matplotlib colormap.

    """
    basename = f'PCA_data_{settings["fol"]}_{n_pca}.hdf'
    data.to_hdf(basename, key='sim_data', mode='w')

    features = data.columns
    data = StandardScaler().fit_transform(data)

    cols = [f'PC{i}' for i in range(1, n_pca + 1)]

    pca_model = PCA(n_components=n_pca)
    principal_comps = pca_model.fit_transform(data)
    principal_df = pd.DataFrame(principal_comps, columns=cols)
    explained = pd.Series(pca_model.explained_variance_ratio_)
    explained.index += 1
    loadings = pd.DataFrame(pca_model.components_.T, columns=cols,
                            index=features)
    load_corr = pca_model.components_.T * np.sqrt(
        pca_model.explained_variance_)
    load_corr_mat = pd.DataFrame(load_corr, columns=cols,
                                 index=features)

    principal_df.to_hdf(basename, key='PC', mode='a')
    loadings.to_hdf(basename, key='loadings', mode='a')
    load_corr_mat.to_hdf(basename, key='correlation_matrix', mode='a')
    explained.to_hdf(basename, key='explained', mode='a')

    pca_info = {'model': pca_model, 'cols': cols, 'features': features}
    _plot_pca_results(n_pca, principal_df, loadings, pca_info, cmap)


def k_means(n_clusters, data, settings, cmap):
    """Perform K-means clustering.

    This function performs K-means clustering on simulation data
    with a chosen amount of clusters, and plots the results.

    Parameters
    ----------
    n_clusters : integer
        Number of clusters.
    data : object like numpy.ndarray
        Simulation data from chosen ensembles.
    settings : dict
        Settings from GUI.
    cmap : string
        Matplotlib colormap.

    """
    model = MiniBatchKMeans(n_clusters)
    labels = model.fit_predict(data)
    fig = plt.figure()
    axis = fig.add_subplot(111)
    axis.scatter(data[:, 0], data[:, 1],
                 c=labels,
                 s=5,
                 cmap=cmap)
    axis.set_title('Kmeans clustering')
    axis.set_xlabel(settings['op1'])
    axis.set_ylabel(settings['op2'])
    plt.show()


def hierarchical(n_clusters, data, settings, cmap):
    """Perform hierarchical clustering.

    This function performs hierarchical clustering on simulation data
    with a chosen amount of clusters, and plots the results.

    Parameters
    ----------
    n_clusters : integer
        Number of clusters.
    data : object like numpy.ndarray
        Simulation data from chosen ensembles.
    settings : dict
        Settings from GUI.
    cmap : string
        Matplotlib colormap.

    """
    model = AgglomerativeClustering(n_clusters=n_clusters,
                                    linkage='average')
    labels = model.fit_predict(data)
    fig = plt.figure()
    axis = fig.add_subplot(111)
    axis.scatter(data[:, 0], data[:, 1],
                 c=labels,
                 s=5,
                 cmap=cmap)
    axis.set_title('Hierarchical clustering')
    axis.set_xlabel(settings['op1'])
    axis.set_ylabel(settings['op2'])
    plt.show()


def spectral(n_clusters, data, settings, cmap):
    """Perform spectral clustering.

    This function performs spectral clustering on simulation data
    with a chosen amount of clusters, and plots the results.

    Parameters
    ----------
    n_clusters : integer
        Number of clusters.
    data : object like numpy.ndarray
        Simulation data from chosen ensembles.
    settings : dict
        Settings from GUI.
    cmap : string
        Matplotlib colormap.

    """
    model = SpectralClustering(
        n_clusters=n_clusters,
        affinity='nearest_neighbors',
        n_neighbors=30,
        assign_labels='kmeans')
    labels = model.fit_predict(data)
    fig = plt.figure()
    axis = fig.add_subplot(111)
    axis.scatter(data[:, 0], data[:, 1],
                 c=labels,
                 s=5,
                 cmap=cmap)
    axis.set_title('Spectral clustering')
    axis.set_xlabel(settings['op1'])
    axis.set_ylabel(settings['op2'])
    plt.show()


def gaussian_mixture(n_clusters, data, settings, cmap):
    """Perform gaussian mixture clustering.

    This function performs gaussian mixture clustering on simulation data
    with a chosen amount of clusters, and plots the results.

    Parameters
    ----------
    n_clusters : integer
        Number of clusters.
    data : object like numpy.ndarray
        Simulation data from chosen ensembles.
    settings : dict
        Settings from GUI.
    cmap : string
        Matplotlib colormap.

    """
    model = GaussianMixture(n_clusters)
    labels = model.fit_predict(data)
    fig = plt.figure()
    axis = fig.add_subplot(111)
    axis.scatter(data[:, 0], data[:, 1],
                 c=labels,
                 s=5,
                 cmap=cmap)
    axis.set_title('Gaussian mixture clustering')
    axis.set_xlabel(settings['op1'])
    axis.set_ylabel(settings['op2'])
    plt.show()


def correlation_matrix(dataframe):
    """Show correlation matrix of simulation variables.

    Function which computes the Pearson correlations between parameters
    and shows it in a correlation plot.

    Parameters
    ----------
    dataframe : object like pandas.DataFrame
        Dataframe containing simulation data.

    """
    titles = list(dataframe.columns)
    fig = plt.figure()
    axis = fig.add_subplot(111)
    corr = axis.matshow(dataframe.corr())
    fig.colorbar(corr)
    axis.set_xticks(np.arange(len(titles)))
    axis.set_yticks(np.arange(len(titles)))
    axis.set_xticklabels(titles)
    axis.set_yticklabels(titles)
    plt.show()


def random_forest(xdata, ydata, depth):
    """Create random forest classification.

    Parameters
    ----------
    xdata : object like pandas.DataFrame
        Pandas dataframe containing the op and cv data from selected
        frames in the simulation.
    ydata : object like pandas.DataFrame
        Pandas dataframe containing True/False for each frame in the
        selected ensembles. True if the frame is reactive, else False.
    depth : integer
        Depth of random forest model.

    """
    rf_model = RandomForestClassifier(n_estimators=500,
                                      max_depth=depth,
                                      random_state=16)
    ydata = ydata.values.ravel()
    rf_model.fit(xdata, ydata)
    important_features = np.argsort(rf_model.feature_importances_)
    feature_imp = pd.Series(rf_model.feature_importances_,
                            index=xdata.columns).sort_values(ascending=False)
    figure = plt.figure()
    axis = figure.add_subplot(111)
    x_ticks = [i + 0.5 for i in range(len(important_features))]
    axis.bar(x_ticks, feature_imp)
    axis.set_xticks(x_ticks)
    axis.set_xticklabels(feature_imp.keys(), rotation=90, ha='center')
    axis.set_ylabel('Gini feature importance')
    plt.show()


def decision_tree(xdata, ydata, depth):
    """Create a decision tree.

    Parameters
    ----------
    xdata : object like pandas.DataFrame
        Pandas dataframe containing the op and cv data from selected
        frames i the simulation.
    ydata : object like pandas.DataFrame
        Pandas dataframe containing True/False for each frame in the
        selected ensembles. True if the frame is reactive, else False.
    depth : integer
        Depth of the decision tree.

    """
    names = xdata.columns.astype(str).tolist()
    x_train, _, y_train, _ = train_test_split(xdata, ydata,
                                              test_size=0.2,
                                              stratify=ydata,
                                              random_state=16,
                                              shuffle=True)

    clf = DecisionTreeClassifier(criterion='entropy', max_depth=depth)
    clf.fit(x_train, y_train)

    dot_data = export_graphviz(clf, out_file=None,
                               feature_names=names,
                               class_names=['reactive', 'unreactive'],
                               filled=True, rounded=True,
                               special_characters=True)
    graph = graphviz.Source(dot_data)
    graph.render('decision_tree', view=True, format='png', cleanup=True)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Simple script to compare the outcome of two simulations.

Here we compare a full simulation with one where we have stopped
and restarted after 100 steps.
"""
import itertools
import os
import sys
import tempfile
import fileinput
import colorama
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import gridspec
from pyretis.inout import print_to_screen
from pyretis.inout.formats.cross import CrossFile
plt.style.use('seaborn-v0_8')


def files_overlap(file1, file2):
    """Check for overlap between two files.

    Here, the overlap is defined as the last line of file1 and
    the first line of file2.

    Parameters
    ----------
    file1 : string
        The file to read the last line from.
    file2 : string
        The file to read the first line from.

    """
    last_line = None
    with open(file1, 'r', encoding='utf-8') as infile1:
        for lines in infile1:
            last_line = lines
    first_line = None
    with open(file2, 'r', encoding='utf-8') as infile2:
        for lines in infile2:
            first_line = lines
            break
    return last_line == first_line


def compare_lists(list1, list2):
    """Compare contents of two lists."""
    if len(list1) == len(list2):
        for i, j in zip(list1, list2):
            if not i == j:
                return 1
        return 0
    return 1


def compare_lists_lists(list1, list2):
    """Compare contents of lists of lists."""
    if len(list1) == len(list2):
        for list1i, list2i in zip(list1, list2):
            if len(list1i) == len(list2i):
                for i, j in zip(list1i, list2i):
                    if not i == j:
                        return 1
            else:
                return 1
        return 0
    return 1


def remove_comments(filename):
    """Remove commented lines if present."""
    for line in fileinput.input(filename, inplace=1):
        if '#' not in line:
            sys.stdout.write(line)


def compare_crossings(file11, file12, file2):
    """Compare the crossings found.

    Here, we expect that the contents of file11 and file12 together
    equals the contents of file2. In some cases, there can be an
    overlap between file11 and file12, and we check for that as well.

    Parameters
    ----------
    file11 : string
        The file containing the first part of the crossing data.
    file12 : string
        The file containing the second part of the crossing data.
    file2 : string
        The file containing the full crossing data.

    """
    remove_comments(file12)
    data1 = []
    with tempfile.NamedTemporaryFile(prefix='cross_', suffix='.txt') as tmp:
        print_to_screen(
            f'Joining: {file11} + {file12} -> {tmp.name}',
            level='info'
        )
        with open(tmp.name, 'w', encoding='utf-8') as outfile:
            with open(file11, 'r', encoding='utf-8') as infile:
                for lines in infile:
                    outfile.write(lines)
            skip = files_overlap(file11, file12)
            with open(file12, 'r', encoding='utf-8') as infile:
                for i, lines in enumerate(infile):
                    if i == 0:
                        if not skip:
                            outfile.write(lines)
                    else:
                        outfile.write(lines)
        data1 = list(CrossFile(tmp.name, 'r').load())
    data2 = list(CrossFile(file2, 'r').load())
    if len(data1) != len(data2):
        print_to_screen('Different number of blocks in crossing data!',
                        level='error')
        return 1
    # Compare blocks
    for block1, block2 in zip(data1, data2):
        ret = compare_lists(block1['comment'], block2['comment'])
        if ret != 0:
            print_to_screen('Comment in crossing files do not match!',
                            level='error')
            return ret
        ret = compare_lists_lists(block1['data'], block2['data'])
        if ret != 0:
            print_to_screen('Crossing files do not match!', level='error')
            print_to_screen(block1['data'], block2['data'])
            return ret
    print_to_screen('Crossing files match!', level='success')
    return 0


def compare_step_output(file11, file12, file2):
    """Compare step-wise output from PyRETIS.

    Here, we assume that the output in file2 is obtained
    by concatenating file11 and file12. file11 and file12
    should have one line in common: the final line in file11
    should be identical to the first line in file12.

    Parameters
    ----------
    file11 : string
        A file to open, part 1.
    file12 : string
        A file to open, part 2.
    file2 : string
        A file to open, we will check if this one is
        equal (line-by-line) to file11+file12.

    Returns
    -------
    None, just prints out the result of the comparison.

    """
    remove_comments(file12)
    print_to_screen('\nComparing files:', level='info')
    print_to_screen(f'{file11} + {file12} == {file2}?')
    with open(file11, 'r', encoding='utf-8') as file11_h:
        with open(file12, 'r', encoding='utf-8') as file12_h:
            with open(file2, 'r', encoding='utf-8') as file2_h:
                first_line = next(file12_h)
                error = False
                for i, (dt1, dt2) in enumerate(zip(itertools.chain(file11_h,
                                                                   file12_h),
                                               file2_h)):
                    if not dt1 == dt2:
                        print_to_screen(f'Error for line no: {i}',
                                        level='error')
                        print_to_screen('Lines were:', level='error')
                        print_to_screen(dt1.strip(), level='error')
                        print_to_screen(dt2.strip(), level='error')
                        error = True
                        break

    if not error:
        print_to_screen('Joined files contain same data!')
        last_line, len11 = read_last_line(file11)
        if last_line == first_line:
            print_to_screen(f'First in {file11} = last in {file12}')
            _, len12 = read_last_line(file12)
            _, len2 = read_last_line(file2)
            if len11 + len12 == len2 + 1:
                print_to_screen('Number of lines are correct.')
                print_to_screen('Files are equal!', level='success')
                return 0
            print_to_screen('Number of lines are incorrect', level='error')
            print_to_screen(len11, len12)
            return 1
        print_to_screen(f'First in {file11} != last in {file12}',
                        level='error')
        return 1
    return 1


def read_last_line(filename):
    """Read the last line from a file + count number of lines in the file."""
    i = 0
    last_line = None
    with open(filename, 'r', encoding='utf-8') as infile:
        for lines in infile:
            last_line = lines
            i += 1
    return last_line, i


def make_fig():
    """Plot for comparison."""
    fig1 = plt.figure(figsize=(12, 6))
    grid = gridspec.GridSpec(2, 2)
    ax1 = fig1.add_subplot(grid[:, 0])
    ax1.plot([], [], label='Potential', lw=0, alpha=0)
    ax1.plot([], [], label='Kinetic', lw=0, alpha=0)
    ax1.plot([], [], label='Total', lw=0, alpha=0)
    ax1.set_xlabel('Step no.')
    ax1.set_ylabel('Energy per particle')
    ax2 = fig1.add_subplot(grid[0, 1])
    ax2.set_ylabel('Temperature')
    ax3 = fig1.add_subplot(grid[1, 1])
    ax3.set_xlabel('Step no.')
    ax3.set_ylabel('Order parameter.')
    ax3.plot([], [], label=r'$\lambda$', lw=0, alpha=0)
    ax3.plot([], [], label=r'$\dot{\lambda}}$', lw=0, alpha=0)
    axes = (ax1, ax2, ax3)
    return fig1, axes


def plot_in_ax(axes, infile, lab, fat=False, colors=None, style='-'):
    """Just do some plotting."""
    ax1, ax2, _ = axes
    data = np.loadtxt(infile)
    if fat:
        width = 7
    else:
        width = 3
    lines = []
    for i, idx in enumerate((1, 2, 3)):
        if colors is None:
            line, = ax1.plot(data[:, 0], data[:, idx], label=lab,
                             ls=style, lw=width, alpha=0.8)
        else:
            line, = ax1.plot(data[:, 0], data[:, idx], label=lab,
                             ls=style, lw=width, alpha=0.8, color=colors[i])
        lines.append(line)
    ax2.plot(data[:, 0], data[:, 4], label=lab, ls=style, lw=width, alpha=0.9)
    return lines


def plot_in_ax_op(axes, infile, lab, fat=False, colors=None, style='-'):
    """Just do some plotting."""
    _, _, ax3 = axes
    data = np.loadtxt(infile)
    if fat:
        width = 7
    else:
        width = 3
    lines = []
    for i, idx in enumerate((1, 2)):
        if colors is None:
            line, = ax3.plot(data[:, 0], data[:, idx], label=lab,
                             ls=style, lw=width, alpha=0.8)
        else:
            line, = ax3.plot(data[:, 0], data[:, idx], label=lab,
                             ls=style, lw=width, alpha=0.8, color=colors[i])
        lines.append(line)
    return lines


def make_plots():
    """Just plot some energies for comparison."""
    figure, axes = make_fig()

    plot_in_ax(
        axes,
        os.path.join('run-full', 'energy.txt'),
        'full',
        fat=True,
        style='-'
    )
    lines = plot_in_ax(
        axes,
        os.path.join('run-step1', 'energy.txt'),
        'restart-part1',
        style='--'
    )
    colors = [i.get_color() for i in lines]
    plot_in_ax(
        axes,
        os.path.join('run-step2', 'energy.txt'),
        'restart-part2',
        style=':', colors=colors
    )

    plot_in_ax_op(
        axes,
        os.path.join('run-full', 'order.txt'),
        'full',
        fat=True,
        style='-'
    )
    lines = plot_in_ax_op(
        axes,
        os.path.join('run-step1', 'order.txt'),
        'restart-part1',
        style='--'
    )
    colors = [i.get_color() for i in lines]
    plot_in_ax_op(
        axes,
        os.path.join('run-step2', 'order.txt'),
        'restart-part2',
        style=':', colors=colors
    )

    axes[0].legend(prop={'size': 'medium'}, ncol=4)
    axes[1].legend(prop={'size': 'medium'})
    axes[2].legend(prop={'size': 'medium'}, ncol=4)
    figure.subplots_adjust(bottom=0.12, right=0.95, top=0.95,
                           left=0.08, wspace=0.2)
    return figure


def main(args):
    """Run comparisons."""
    result = 0
    ret = compare_crossings(
        os.path.join('run-step1', 'cross.txt'),
        os.path.join('run-step2', 'cross.txt'),
        os.path.join('run-full', 'cross.txt'),
    )
    result += ret

    ret = compare_step_output(
        os.path.join('run-step1', 'energy.txt'),
        os.path.join('run-step2', 'energy.txt'),
        os.path.join('run-full', 'energy.txt'),
    )
    result += ret

    ret = compare_step_output(
        os.path.join('run-step1', 'order.txt'),
        os.path.join('run-step2', 'order.txt'),
        os.path.join('run-full', 'order.txt'),
    )
    result += ret
    if 'make_plot' in args:
        fig = make_plots()
        fig.savefig('compare.png')
        plt.show()
    if result != 0:
        print_to_screen('\nComparison failed!', level='error')
    else:
        print_to_screen('\nComparison was successful!', level='success')
    return result


if __name__ == '__main__':
    colorama.init(autoreset=True)
    sys.exit(main(sys.argv[1:]))



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""The Langevin integrator implemented in FORTRAN."""
import logging
import numpy as np
from matplotlib import pyplot as plt
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())
try:
    from vvintegrator import vvintegrator
except ImportError:
    MSG = ('Could not import external FORTRAN library.'
           '\nPlease compile with "make"!')
    logger.critical(MSG)
    raise ImportError(MSG)


plt.style.use('seaborn-colorblind')
SEED = 1000


def test_rangaussian(sigma, numbers=10000, bins=100):
    """Test the gaussian generator."""
    rndf = np.array([vvintegrator.rangaussian(sigma) for _ in range(numbers)])
    rnd = np.random.normal(loc=0.0, scale=sigma, size=numbers)

    mini = min(rndf.min(), rnd.min())
    maxi = max(rndf.max(), rnd.max())

    fig = plt.figure()
    ax1 = fig.add_subplot(111)

    histf, edgesf = np.histogram(rndf, bins=bins, range=(mini, maxi))
    midf = 0.5 * (edgesf[1:] + edgesf[:-1])
    hist, edges = np.histogram(rnd, bins=bins, range=(mini, maxi))
    mid = 0.5 * (edges[1:] + edges[:-1])
    ax1.plot(midf, histf, 'o-', label='FORTRAN', lw=2, alpha=0.7)
    ax1.plot(mid, hist, 's-', label='numpy', lw=2, alpha=0.7)
    ax1.set_title(f'Numbers = {numbers}, bins = {bins}')
    ax1.legend()
    plt.show()


def test_gssbivar(s12os11, sqrts11, sqrtsos11, numbers=10000, bins=100):
    """Test the gaussian bivariate distribution."""
    rndf = []
    for _ in range(numbers):
        rndf.append(vvintegrator.gssbivar(s12os11, sqrts11, sqrtsos11))
    rndf = np.array(rndf)
    histf, xedges, yedges = np.histogram2d(rndf[:, 0], rndf[:, 1],
                                           bins=bins, range=None)
    xvarf, yvarf = np.meshgrid(xedges, yedges)
    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.pcolormesh(xvarf, yvarf, histf)
    plt.show()


def main():
    """Run the test."""
    size = vvintegrator.get_seed_size()
    seeds = np.array([SEED + i for i in range(size)], dtype='i4')
    vvintegrator.seed_random_generator(seeds)
    for sig in [1.0, 0.5, 2.0, 10.0]:
        test_rangaussian(sig, numbers=10**6)
    test_gssbivar(1.0, 2.0, 1.0)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the lattice generation tools in pyretis.tools"""
import logging
import unittest
import numpy as np
from pyretis.tools.lattice import generate_lattice
logging.disable(logging.CRITICAL)


class LatticeTest(unittest.TestCase):
    """Test the lattice generation in pyretis.tools.lattice."""

    def test_lattice_generation(self):
        """Test that we can generate all lattices."""
        xyz = {'fcc': np.array([[0., 0., 0.], [1., 1., 0.],
                                [0., 1., 1.], [1., 0., 1.]]),
               'bcc': np.array([[0., 0., 0.], [0.5, 0.5, 0.5],
                                [1., 0., 0.], [1.5, 0.5, 0.5]]),
               'diamond': np.array([[0., 0., 0.],
                                    [0., 0.705, 0.705],
                                    [0.705, 0., 0.705],
                                    [0.705, 0.705, 0.],
                                    [0.3525, 0.3525, 0.3525],
                                    [0.3525, 1.0575, 1.0575],
                                    [1.0575, 0.3525, 1.0575],
                                    [1.0575, 1.0575, 0.3525]]),
               'hcp': np.array([[0., 0., 0.], [0.45, 0.45, 0.],
                                [0.45, 0.75, 0.45], [0., 0.3, 0.45]]),
               'sc': np.array([[0., 0., 0.],
                               [0., 0., 3.14159],
                               [0., 0., 6.28318],
                               [0., 0., 9.42477]]),
               'sq': np.array([[0., 0.], [0., 3.], [3., 0.], [3., 3.]]),
               'sq2': np.array([[0., 0.], [1.5, 1.5], [0., 3.], [1.5, 4.5],
                                [3., 0.], [4.5, 1.5], [3., 3.], [4.5, 4.5]])}

        size = {'fcc': np.array([[0.0, 2.0], [0.0, 2.0], [0.0, 2.0]]),
                'bcc': np.array([[0.0, 2.0], [0.0, 1.0], [0.0, 1.0]]),
                'diamond': np.array([[0.0, 1.41], [0.0, 1.41], [0.0, 1.41]]),
                'hcp': np.array([[0.0, 0.9], [0.0, 0.9], [0.0, 0.9]]),
                'sc': np.array([[0.0, 3.14159], [0.0, 3.14159],
                                [0.0, 12.56636]]),
                'sq': np.array([[0.0, 6], [0.0, 6]]),
                'sq2': np.array([[0.0, 6], [0.0, 6]])}

        args = {'bcc': [2, 1, 1],
                'fcc': [1, 1, 1],
                'hcp': [1, 1, 1],
                'diamond': [1, 1, 1],
                'sc': [1, 1, 4],
                'sq': [2, 2],
                'sq2': [2, 2]}

        kwargs = {'bcc': dict(lcon=1.0),
                  'fcc': dict(lcon=2.0),
                  'hcp': dict(lcon=0.9),
                  'sc': dict(lcon=3.14159),
                  'sq': dict(lcon=3),
                  'sq2': dict(lcon=3),
                  'diamond': dict(lcon=1.41)}
        # Test bcc:
        for key in args:
            xyz_g, size_g = generate_lattice(key, args[key], **kwargs[key])
            self.assertTrue(np.allclose(xyz_g, xyz[key], atol=1.0e-8))
            self.assertTrue(np.allclose(np.array(size_g), size[key],
                                        atol=1.0e-8))
        # Test if we give almost nothing
        pos, size = generate_lattice(lattice='1d')
        self.assertEqual(pos[0], 0)
        self.assertEqual(size, [[0.0, 1]])

    def test_generate_no_lcon_or_dens(self):
        """Test if the generate lattice methods fails when it should."""
        # Test if we feed it some unknown lattice:
        args = ['undefined lattice', None]
        kwargs = dict(lcon=None, density=None)
        self.assertRaises(ValueError, generate_lattice,
                          *args, **kwargs)
        # Test if we feed it some strange arguments:
        args = [['undefined', 'lattice'], None]
        self.assertRaises(ValueError, generate_lattice,
                          *args, **kwargs)
        # Test if we do not give lattice parameters:
        args = ['sc', [1, 1, 1]]
        self.assertRaises(ValueError, generate_lattice, *args)
        # Test if we give to few repeats
        args = ['sc', [1, 1]]
        kwargs = dict(lcon=1.0)
        self.assertRaises(ValueError, generate_lattice, *args, **kwargs)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A template for a new external MD engine."""
import logging
import os
from pyretis.engines.external import ExternalMDEngine


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

INPUT_CONF = []
OUTPUT_FILES = []
REQUIRED_FILES = []


def get_energy():
    """Fictional function to read energy."""
    return 42, 42


def kinetic_energy_output_from_velocity_generation(_, __):
    """Fictional function to read kin eng."""
    return 42


def method_to_read_config_file(_):
    """Fictional function to read config file."""
    return 42, 42


def method_to_write_config_file(_, __, ___):
    """Fictional function to write a config file."""
    return


class NewEngine(ExternalMDEngine):
    """A new external engine class.

    Attributes
    ----------
    exe : string
        The command for executing the engine.
    input_path : string
        The directory where the input files are stored.
    timestep : float
        The time step used in the external MD simulation.
    subcycles : integer
        The number of steps each external step is composed of.
    """

    def __init__(self, exe, input_path, timestep, subcycles):
        """Initiate the script.

        Parameters
        ----------
        exe : string
            The engine executable.
        input_path : string
            The absolute path to where the input files are stored.
        timestep : float
            The time step used in the external engine.
        subcycles : integer
            The number of steps each external step is composed of.
        """
        super().__init__('New external engine', timestep, subcycles)
        # Store variables for the engine:
        self.exe = exe
        # store input path:
        self.input_path = os.path.abspath(input_path)
        # store input files:
        self.input_files = {}
        for key in REQUIRED_FILES:
            self.input_files[key] = os.path.join(self.input_path, key)
            if not os.path.isfile(self.input_files[key]):
                msg = 'Engine could not find file "{}"!'.format(key)
                raise ValueError(msg)

        # Create a pyretis version of the input file:
        settings = {'timestep': self.timestep,
                    'subcycles': self.subcycles}

        self.input_files['input'] = os.path.join(self.input_path,
                                                 'PYRETIS_INPUT')
        self._modify_input(self.input_files['ORIGINAL_INPUT'],
                           self.input_files['input'], settings, delim='=')

    def _run_program(self):
        """Method to execute the new engine.

        Returns
        -------
        out : dict
            The files created by the run.
        """
        self.execute_command(self.exe, cwd=self.exe_dir, inputs=None)
        out = {}
        for key in OUTPUT_FILES:
            filename = os.path.join(self.exe_dir, key)
            if os.path.isfile(filename):
                out[key] = filename
        return out

    def _extract_frame(self, traj_file, idx, out_file):
        """Extract a frame from a trajectory file.

        This method is used by `self.dump_config` when we are
        dumping from a trajectory file. It is not used if we are
        dumping from a single config file. Here you can write
        your own code for dumping or make use of existing libraries
        such as mdtraj, mdanalysis, ase, ...

        Parameters
        ----------
        traj_file : string
            The trajectory file to dump from.
        idx : integer
            The frame number we look for.
        out_file : string
            The file to dump to.
        """
        return

    def _propagate_from(self, name, path, ensemble, reverse=False):
        """Propagate from the current system configuration.

        Here, we assume that this method is called after the propagate()
        has been called in the parent. The parent is then responsible
        for reversing the velocities and also for setting the initial
        state of the system.

        Parameters
        ----------
        name : string
            A name to use for the trajectory we are generating.
        path : object like :py:class:`.PathBase`
            This is the path we use to fill in phase-space points.
            We are here not returning a new path - this since we want
            to delegate the creation of the path to the method
            that is running `propagate`.
        ensemble: dict
            it contains:
            * `system` : object like :py:class:`.System`
              The system object gives the initial state for the
              integration. The initial state is stored and the system is
              reset to the initial state when the integration is done.
            * `order_function` : object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.
            * `interfaces` : list of floats
              These interfaces define the stopping criterion.
        reverse : boolean
            If True, the system will be propagated backwards in time.

        Returns
        -------
        success : boolean
            This is True if we generated an acceptable path.
        status : string
            A text description of the current status of the propagation.
        """
        # Dumping of the initial config were done by the parent, here
        # we will just the dumped file:
        initial_conf = ensemble['system'].particles.get_pos()[0]
        # Copy this file, in case the external program requires
        # specific names for the input configurations:
        input_conf = os.path.join(self.exe_dir, 'input_conf')
        output_conf = os.path.join(self.exe_dir, 'output_conf')
        self._copyfile(initial_conf, output_conf)
        # Set the positions to be just the file we copied:
        ensemble['system'].particles.set_pos((input_conf, None))
        # Create the output trajectory file:
        traj_file = os.path.join(self.exe_dir, name)
        # Add other files here as well, input settings, force field etc...
        success = False
        left, _, right = ensemble['interfaces']
        phase_point = ensemble['system'].particles.get_particle_state()
        # Just to be keep the same order as in the loop
        vpot = phase_point['vpot']
        ekin = phase_point['ekin']
        order = self.calculate_order(ensemble)
        for i in range(path.maxlen):
            # We first add the current phase point, and then we propagate.
            phase_point = {
                'order': order,
                'pos': (traj_file, i),
                'vel': reverse,
                'vpot': vpot,
                'ekin': ekin}
            status, success, stop = self.add_to_path(path, phase_point,
                                                     left, right)
            # Add the current frame to the output trajectory also:
            # e.g.: add_frame_to_traj('CURRENT_FRAME_FILE', traj_file, i)
            if stop:
                logger.debug('Ending propagate at %i. Reason: %s', i, status)
                break
            # Here execute the external program
            # for instance by using self.step() or self._run_program()
            # get energies after execution:
            ekin, vpot = get_energy()
            # Recalculate order parameter
            # If the output file is named different, update the
            # system object so that it points to this file, e.g.:
            # system.particles.set_pos((NAME_OF_FILE, None)), or
            # if the energies are needed as well, use
            # system.particles.set_particle_state(...)
            order = self.calculate_order(ensemble)
        # We are done with the execution:
        # clean up and remove files:
        list_of_files_to_remove = []
        self._remove_files(self.exe_dir, list_of_files_to_remove)
        return success, status

    def step(self, system, name):
        """Perform a single step with the external engine.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we are integrating.
        name : string
            To name the output files from the step.

        Returns
        -------
        out_files : dict
            The output files created by this step.
        """
        # Get initial config:
        input_file = self.dump_frame(system)
        # copy input files, maybe rename input_file etc.
        # Then, run it
        out_files = self._run_program()
        # Store the frame after the step, make use of the name variable:
        config = os.path.join(self.exe_dir, 'OUTPUT.{}'.format(name))
        # Get energies
        ekin, vpot = get_energy()
        # remove files that we dont need after the step:
        files_to_remove = []
        self._remove_files(self.exe_dir, files_to_remove)
        # Update the state of the system to match the current config
        phase_point = {'pos': (config, None),
                       'vel': False, 'vpot': vpot, 'ekin': ekin}
        system.particles.set_particle_state(phase_point)
        return config

    @staticmethod
    def _read_configuration(filename):
        """Method to read config files for the external engine.

        This method is used when we calculate the order parameter.

        Parameters
        ----------
        filename : string
            The file to read the configuration from.

        Returns
        -------
        xyz : numpy.array
            The positions.
        vel : numpy.array
            The velocities.
        """
        xyz, vel = method_to_read_config_file(filename)
        return xyz, vel

    @staticmethod
    def _reverse_velocities(filename, outfile):
        """Method to reverse velocity in a given snapshot.

        Parameters
        ----------
        filename : string
            The configuration to reverse velocities in.
        outfile : string
            The output file for storing the configuration with
            reversed velocities.
        """
        xyz, vel = method_to_read_config_file(filename)
        method_to_write_config_file(outfile, xyz, -vel)

    def modify_velocities(self, system, rgen, vel_settings):
        """Modify the velocities of the current state.

        This method will modify the velocities of a time slice.

        Parameters
        ----------
        system : object like :py:class:`.System`
            System is used here since we need access to the particle
            list.
        rgen : object like :py:class:`.RandomGenerator`
            This is the random generator that will be used.
        vel_settings: dict
            It contains all the info for the velocity:

            * `sigma_v` : numpy.array, optional
              These values can be used to set a standard deviation (one
              for each particle) for the generated velocities.
            * `aimless` : boolean, optional
              Determines if we should do aimless shooting or not.
            * `momentum` : boolean, optional
              If True, we reset the linear momentum to zero after
              generating.
            * `rescale` : float, optional
              In some NVE simulations, we may wish to re-scale the
              energy to a fixed value. If `rescale` is a float > 0,
              we will re-scale the energy (after modification of
              the velocities) to match the given float.

        Returns
        -------
        dek : float
            The change in the kinetic energy.
        kin_new : float
            The new kinetic energy.
        """
        dek = None
        kin_old = None
        kin_new = None
        rescale = vel_settings.get('rescale')

        if rescale is not None and rescale is not False and rescale > 0:
            # Code to support energy rescale, or just ignore this as
            # set:
            msgtxt = 'External engine does not support energy rescale!'
            logger.error(msgtxt)
            raise NotImplementedError(msgtxt)
        else:
            kin_old = system.particles.ekin
        if vel_settings.get('aimless', False):
            # pos = self.dump_frame(system)
            phase_point = system.particles.get_particle_state()
            # Generate new velocities, either by hand or by asking the
            # external program.
            genvel = os.path.join(self.exe_dir, 'g_POSCAR')
            # Also get the new kinetic energy
            kin_new = kinetic_energy_output_from_velocity_generation(genvel,
                                                                     rgen)
            # Update the phase point (note we keep the potential energy):
            phase_point['pos'] = (genvel, None)
            phase_point['vel'] = False
            phase_point['ekin'] = kin_new
            system.particles.set_particle_state(phase_point)
        else:  # soft velocity change, add from Gaussian dist
            # You might not need to support this:
            msgtxt = 'External engine only support aimless shooting!'
            logger.error(msgtxt)
            raise NotImplementedError(msgtxt)
        if kin_old is None or kin_new is None:
            dek = float('inf')
            logger.warning(('Kinetic energy not found for previous point.'
                            '\n(This happens when the initial configuration '
                            'does not contain energies.)'))
        else:
            dek = kin_new - kin_old
        return dek, kin_new



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Create a directive for showing input examples."""
from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import unchanged
from jinja2 import Template


HTML_PRE = """
<div class="container">
    <div class="panel panel-keyword">
      <div class="panel-heading">{{ heading }}</div>
      <div class="panel-body">"""
TEMPLATE_PRE = Template(HTML_PRE)
HTML_POST = """</div>
  </div>
</div>"""
TEMPLATE_POST = Template(HTML_POST)


class PyretisExampleNode(nodes.General, nodes.Element):
    pass


class PyretisInputExample(Directive):
    """Create a new directive for PyRETIS examples."""

    required_arguments = 0
    optional_arguments = 1
    option_spec = {'class-name': unchanged}
    has_content = True

    def run(self):
        node = PyretisExampleNode()
        class_name = self.options.get('class-name', None)
        if not self.arguments:
            section = 'Example section'
        else:
            section = 'Example {} section'.format(self.arguments[0])
        if class_name:
            node['heading'] = '{} for {}:'.format(section, class_name)
        else:
            node['heading'] = '{}:'.format(section)
        self.state.nested_parse(self.content, self.content_offset, node)
        return [node]


def html_visit_input_example_block(self, node):
    self.body.append(
        TEMPLATE_PRE.render(
            heading=node['heading']
        )
    )


def html_depart_input_example_block(self, node):
    self.body.append(TEMPLATE_POST.render())


def setup(app):
    app.add_node(
        PyretisExampleNode,
        html=(
            html_visit_input_example_block,
            html_depart_input_example_block
        )
    )
    app.add_directive('pyretis-input-example', PyretisInputExample)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the system class from pyretis.core.system"""
import logging
import unittest
import numpy as np
from pyretis.core.box import create_box
from pyretis.core.particles import Particles
from pyretis.core.system import System
from pyretis.core.units import UNIT_SYSTEMS, CONSTANTS
from pyretis.forcefield import ForceField, PotentialFunction


logging.disable(logging.CRITICAL)


class FakePotential(PotentialFunction):
    """A potential which may potentially be used in testing."""

    def __init__(self, desc='Test potential'):
        """Initiate."""
        super().__init__(dim=3, desc=desc)
        self.pstate = 10
        self.fstate = 10

    def potential(self, system):
        """Evaluate the fake potential."""
        self.pstate -= 1 * system.particles.npart
        return self.pstate

    def force(self, system):
        """Evaluate the fake force."""
        pos = system.particles.pos
        self.fstate -= 1
        vir = np.eye(3, 3) * self.fstate
        return self.fstate * np.ones_like(pos), vir

    def potential_and_force(self, system):
        """Evaluate potential and force."""
        pot = self.potential(system)
        force, virial = self.force(system)
        return pot, force, virial


def prepare_test_system():
    """Create system with some particles."""
    box = create_box(cell=[10., 10., 10])
    particles = Particles(dim=3)
    particles.add_particle(np.zeros(3), np.zeros(3), np.zeros(3))
    particles.add_particle(np.zeros(3), np.zeros(3), np.zeros(3))
    particles.add_particle(np.zeros(3), np.zeros(3), np.zeros(3))
    syst = System(box=box, temperature=10)
    syst.particles = particles
    forcefield = ForceField('Testing force field')
    pot = FakePotential()
    forcefield.add_potential(pot)
    syst.forcefield = forcefield
    return syst


class SystemTest(unittest.TestCase):
    """Run the tests for the System class."""

    def test_adjust_dof(self):
        """Test if we can change the dofs."""
        system = System()
        system.adjust_dof([1, 1, 1])
        self.assertTrue(np.allclose(system.temperature['dof'],
                                    [1, 1, 1]))
        system.adjust_dof([0, 0, 9])
        self.assertTrue(np.allclose(system.temperature['dof'],
                                    [1, 1, 10]))

    def test_get_boltzmann(self):
        """Test that kB is set correctly."""
        for key in UNIT_SYSTEMS:
            syst = System(units=key)
            self.assertAlmostEqual(syst.get_boltzmann(),
                                   CONSTANTS['kB'][key])

    def test_restart_info(self):
        """Test that we can create restart info."""
        box = create_box(cell=[1, 2, 3])
        syst = System(units='lj', box=box, temperature=10.)
        restart = syst.restart_info()
        correct = {
            'temperature': {'set': 10.0, 'dof': np.array([1, 1, 1]),
                            'beta': 0.1},
            'box': {'high': np.array([1., 2., 3.]),
                    'periodic': [True, True, True],
                    'low': np.array([0., 0., 0.]),
                    'length': np.array([1., 2., 3.])},
            'units': 'lj'}
        for key in correct:
            self.assertTrue(key in restart)
        self.assertEqual(correct['units'], restart['units'])
        for key1 in ('temperature', 'box'):
            for key in correct[key1]:
                self.assertTrue(np.allclose(correct[key1][key],
                                            restart[key1][key]))

        syst = System(units='real', box=None, temperature=100.)
        restart = syst.restart_info()
        self.assertEqual('real', restart['units'])
        self.assertTrue(restart['temperature']['dof'] is None)
        self.assertAlmostEqual(restart['temperature']['beta'],
                               1./(100 * CONSTANTS['kB']['real']))
        self.assertAlmostEqual(restart['temperature']['set'], 100.)

    def test_evaluate_fp(self):
        """Test that we can evaluate forces and potential."""
        syst = prepare_test_system()
        pos = syst.particles.pos
        self.assertEqual(syst.potential(), 7)
        self.assertEqual(syst.evaluate_potential(), 4)
        pot, force, virial = syst.evaluate_potential_and_force()
        self.assertEqual(pot, 1)
        self.assertTrue(np.allclose(force, 9 * np.ones_like(pos)))
        self.assertTrue(np.allclose(virial, 9 * np.eye(3, 3)))
        force, virial = syst.force()
        self.assertTrue(np.allclose(force, 8 * np.ones_like(pos)))
        self.assertTrue(np.allclose(virial, 8 * np.eye(3, 3)))
        force, virial = syst.evaluate_force()
        self.assertTrue(np.allclose(force, 7 * np.ones_like(pos)))
        self.assertTrue(np.allclose(virial, 7 * np.eye(3, 3)))

    def test_generate_vel(self):
        """Test that we can generate velocities."""
        syst = prepare_test_system()
        vel = syst.particles.vel
        syst.generate_velocities(rgen='mock', seed=0, momentum=True,
                                 temperature=None, distribution='maxwell')
        correct_vel = np.array([[0.61114808, -2.78504494, 1.42713429],
                                [2.25062036, 1.32763469, 3.26596079],
                                [-2.86176844, 1.45741024, -4.69309508]])
        self.assertTrue(np.allclose(vel, correct_vel))
        logging.disable(logging.INFO)
        with self.assertLogs('pyretis.core.system', level='ERROR'):
            with self.assertLogs('pyretis.core.random_gen', level='CRITICAL'):
                syst.generate_velocities(rgen='mock',
                                         distribution='fake news')
        logging.disable(logging.CRITICAL)

    def test_calculate_temp(self):
        """Test that we can calculate temperatures."""
        syst = prepare_test_system()
        syst.generate_velocities(rgen='mock', seed=0, momentum=True,
                                 temperature=None, distribution='maxwell')
        temp = syst.calculate_temperature()
        self.assertAlmostEqual(temp, 10.)

    def test_rescale_velocities(self):
        """Test that we can rescale velocities."""
        syst = prepare_test_system()
        syst.generate_velocities(rgen='mock', seed=0, momentum=True,
                                 temperature=None, distribution='maxwell')
        syst.rescale_velocities(13)
        temp = syst.calculate_temperature()
        self.assertAlmostEqual(temp, 2.)
        logging.disable(logging.INFO)
        with self.assertLogs('pyretis.core.system', level='WARNING'):
            syst.rescale_velocities(1)
        logging.disable(logging.CRITICAL)

    def test_extra_setup(self):
        """Test that we can do extra set-up."""
        syst = prepare_test_system()
        syst.generate_velocities(rgen='mock', seed=0, momentum=True,
                                 temperature=None, distribution='maxwell')
        temp = syst.calculate_temperature()
        self.assertAlmostEqual(temp, 10.)
        syst.extra_setup()

        syst.post_setup.append(('rescale_velocities', [11]))
        syst.extra_setup()
        temp = syst.calculate_temperature()
        self.assertAlmostEqual(temp, 1.333333333)

    def test_update_box(self):
        """Test that we can update the system box."""
        syst = System()
        length = [1, 2, 3, 4, 5, 6]
        syst.update_box(length)
        for i, j in zip(syst.box.cell, length):
            self.assertEqual(i, j)
        length2 = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        syst.update_box(length2)
        for i, j in zip(syst.box.cell, length):
            self.assertEqual(i, j)

    def test_system_copy(self):
        """Test that we can copy a system."""
        system = prepare_test_system()
        system.order = 1234
        system_copy = system.copy()
        self.assertIsNot(system, system_copy)
        self.assertEqual(system, system_copy)
        # Check that order parameter was also copied:
        self.assertEqual(system_copy.order, system.order)
        # Test that a change in one of the systems does not alter the other:
        self.assertEqual(system.temperature['set'],
                         system_copy.temperature['set'])
        system.temperature['set'] = 11
        self.assertNotEqual(system.temperature['set'],
                            system_copy.temperature['set'])
        self.assertNotEqual(system, system_copy)
        system.temperature['dof'] = np.array([2, 2, 2])
        # Test what happens is some attribute is missing:
        del system.temperature
        system_copy = system.copy()
        # Systems should not be equal in this case since one is
        # missing an essential attribute:
        self.assertNotEqual(system, system_copy)
        # Test if particles/box are None:
        system = System()
        system_copy = system.copy()
        self.assertTrue(system == system_copy)
        self.assertFalse(system != system_copy)
        # Test if articles/box are missing:
        del system.box
        del system.particles
        with self.assertRaises(AttributeError):
            system.copy()


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module defining functions for analysis of order parameters.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

analyse_orderp (:py:func:`.analyse_orderp`)
    Run a simple order parameter analysis.
"""
from pyretis.analysis.analysis import analyse_data, mean_square_displacement


__all__ = ['analyse_orderp']


def analyse_orderp(orderdata, settings):
    """Run the analysis on several order parameters.

    The results are collected into a structure which is convenient for
    plotting.

    Parameters
    ----------
    orderdata : numpy.arrays
        The data read from the order parameter file.
    settings : dict
        This dictionary contains settings for the analysis.

    Returns
    -------
    results : numpy.array
        For each order parameter `key`, `results[key]` contains the
        result of the analysis.

    See Also
    --------
    `analyse_data` in `pyretis.analysis.analysis.py`

    Note
    ----
    We here (and in the plotting routines) make certain assumptions
    about the structure, i.e. the positions are assumed to have a
    specific meaning: column zero is the time, column one the order
    parameter and so on.

    """
    results = []
    _, col = orderdata.shape
    for i in range(col):
        if i == 0:  # first column is just the time, skip it
            pass
        else:
            result = analyse_data(orderdata[:, i], settings)
            if i == 1:  # assume that we want the MSD analysis here:
                ndt = settings['analysis']['maxordermsd']
                result['msd'] = mean_square_displacement(
                    orderdata[:, i],
                    ndt=ndt
                )
            results.append(result)
    return results



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains classes to represent angle order parameters.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Angle (:py:class:`.Angle`)
    An angle defined by three atoms.
"""
import logging
from numpy import dot
from numpy import arccos  # pylint: disable=no-name-in-module
from numpy import clip
from numpy.linalg import norm
from pyretis.orderparameter.orderparameter import OrderParameter
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['Angle']


class Angle(OrderParameter):
    """An angle order parameter.

    This class defines an order parameter which is an angle
    ABC for 3 particles A, B and C. The angle is defined as the
    angle given by the two vectors BA and BC.

    Attributes
    ----------
    index : list of integers
        These are the indices of atoms to be used for the angle,
        i.e. system.particles.pos[index] will be used.
    periodic : boolean
        This determines if periodic boundaries should be applied to
        the positions/distances.

    """

    def __init__(self, index, periodic=False):
        """Initialise the order parameter.

        Parameters
        ----------
        index : list/tuple of integers
            The indices for the atoms defining the angle.
        periodic : boolean, optional
            This determines if periodic boundary conditions should be
            applied to the distance vectors.

        """
        try:
            if len(index) != 3:
                msg = ('Wrong number of atoms for angle definition. '
                       f'Expected 3 got {len(index)}')
                logger.error(msg)
                raise ValueError(msg)
        except TypeError as err:
            msg = 'Angle should be defined as a tuple/list of integers!'
            logger.error(msg)
            raise TypeError(msg) from err
        txt = f'Angle between particles {index[0]}, {index[1]} and {index[2]}'
        super().__init__(description=txt)
        self.periodic = periodic
        self.index = [int(i) for i in index]

    def calculate(self, system):
        """Calculate the angle.

        Parameters
        ----------
        system : object like :py:class:`.System`
            Object containing the positions and box info we use for
            the calculation.

        Returns
        -------
        out : list of floats
            The order parameters.

        """
        pos = system.particles.pos
        vector_ba = pos[self.index[1]] - pos[self.index[0]]
        vector_bc = pos[self.index[1]] - pos[self.index[2]]
        if self.periodic:
            vector_ba = system.box.pbc_dist_coordinate(vector_ba)
            vector_bc = system.box.pbc_dist_coordinate(vector_bc)
        vector_ba /= norm(vector_ba)
        vector_bc /= norm(vector_bc)
        angle = arccos(clip(dot(vector_ba, vector_bc), -1, 1))
        return [angle]



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""The sub-package with tools for visualizing simulation results for PyRETIS.

This package is intended for compiling data of a simulation into a compact file
standard (hdf5), and displaying results
from file in a custom GUI applet.
Included are compiler of simulation data and a custom-built PyQt5 GUI applet
that loads pre-compiled data (or compiles when loading raw simulation data).
The applet allows for user-friendly and interactive plotting of combinations of
order parameter data of different interfaces and cycles of the simulation.

Package structure
~~~~~~~~~~~~~~~~~

Modules
~~~~~~~

__init__.py
    Imports from the other modules.

common.py (:py:mod:`pyretis.pyvisa.common`)
    Common functions and variables for the visualization. These functions
    are mainly intended for internal use and are not imported here.

orderparam_density.py (:py:mod:`pyretis.pyvisa.orderparam_density`)
    A module that handles the compiling of data to a single file.

plotting.py (:py:mod:`pyretis.pyvisa.plotting`)
    A module which contains some functions that are used to plot regression
    lines and interface planes, and generate surface plots.

resources_rc.py (:py:mod:`pyretis.pyvisa.resources_rc`)
    A module containing the resources, icons/logos for the PyVISA GUI.

visualize.py (:py:mod:`pyretis.pyvisa.visualize`)
    A module that handles the loading and plotting of data from a compiled file
    or a simulation.

Sub-packages
~~~~~~~~~~~~

None

Important classes defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CustomFigCanvas
    (:py:class:`pyretis.pyvisa.visualize.CustomFigCanvas`)
    A class for the custom figure shown in the VisualApp class PyQt5 applet.

DataObject
    (:py:class:`pyretis.pyvisa.visualize.DataObject`)
    A class that reads from simulation data, holds the data, and supplies the
    data to VisualApp for plotting.

DataSlave (:py:class:`pyretis.pyvisa.visualize.DataSlave`)
    QObject class definition that holds the PathDensity data.

PathDensity (:py:class:`pyretis.pyvisa.orderparam_density.PathDensity`)
    A class for reading, storing, and compiling simulation data.

PathVisualize
    (:py:class:`pyretis.pyvisa.orderparam_density.PathVisualize`)
    A class for loading data (compiled or not), and generating plots.

VisualApp (:py:class:`pyretis.pyvisa.visualize.VisualApp`)
    A QtWidget class that holds a user-defined figure.

VisualObject
    (:py:class:`pyretis.pyvisa.visualize.VisualObject`)
    A class that loads from hdf5, holds and, supplies VisualApp with data for
    plotting.


Important methods defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

_grid_it_up (:py:func:`pyretis.pyvisa.plotting._grid_it_up`)
    Maps the x,y and z data to a numpy.meshgrid using scipy interpolation
    at a user defined resolution.

gen_surface(:py:func:`pyretis.pyvisa.plotting.gen_surface`)
    A function that generates a user-defined `surface` plot (2D/3D).

_grid_it_up(:py:func:`pyretis.pyvisa.plotting._grid_it_up`)
    A function that generates a [X,Y] numpy.meshgrid and [Z] grid-data
    for a given resolution.

plot_int_plane(:py:func:`pyretis.pyvisa.plotting.plot_int_plane`)
    A function that generates planes of the simulation interfaces for
    3D plots.

plot_regline(:py:func:`pyretis.pyvisa.plotting.plot_regline`)
    A function that generates a linear regression line of x and y data on
    a given matplotlib.axes object.

shift_data(:py:func:`pyretis.pyvisa.common.shift_data`)
    A function that shifts data values of a list by the median value.

try_data_shift(:py:func:`pyretis.pyvisa.common.try_data_shift`)
    A function that attempts a shift of the data values to increase linear
    correlation.
"""
from .common import (try_data_shift, shift_data)
from .orderparam_density import (PathDensity, PathVisualize)
from .plotting import (plot_regline,
                       _grid_it_up,
                       plot_int_plane,
                       gen_surface)
# Check if PyVisA requirements are installed
try:  # pragma: no cover
    import PyQt5
    import mdtraj
    import sklearn
    from pyretis.pyvisa.visualize import VisualApp
    HAS_PYVISA_REQ = True
except ImportError:
    HAS_PYVISA_REQ = False



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Define the class for a generic potential function.

This module defines the generic class for potential functions.
This class is sub-classed in all potential functions.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

PotentialFunction (:py:class:`.PotentialFunction`)
    A class for representing generic potential functions.
"""
import logging
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['PotentialFunction']


class PotentialFunction:
    """Base class for a generic potential function.

    Generic class for potential functions.

    Attributes
    ----------
    desc : string
        Short description of the potential.
    dim : int
        Represents the spatial dimensionality of the potential.
    params : dict
        The parameters for the potential. This dict defines,
        on initiation, the parameters the potential will handle
        and store.

    """

    def __init__(self, dim=1, desc=''):
        """Initialise the potential.

        Parameters
        ----------
        dim : int, optional
            Represents the dimensionality.
        desc : string, optional
            Description of the potential function. Used to print out
            information about the potential.

        """
        self.dim = dim
        self.desc = desc
        self.params = {}

    def set_parameters(self, parameters):
        """Update all parameters. Input is assumed to be a dict."""
        for key in parameters:
            if key in self.params:
                self.params[key] = parameters[key]
            else:
                msg = 'Could not find "{}" in parameters. Ignoring!'
                msg = msg.format(key)
                logger.warning(msg)
        self.check_parameters()

    def check_parameters(self):
        """Check the consistency of the parameters.

        Returns
        -------
        out : boolean
            True if the check(s) pass.

        """
        if not self.params:
            logger.warning('No parameters are set for the potential')
            return False
        return True

    def __str__(self):
        """Return the string description of the potential."""
        msg = [f'Potential: {self.desc}']
        strinfo = '{}: {}'
        for key in sorted(self.params):
            msg.append(strinfo.format(key, self.params[key]))
        return '\n'.join(msg)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module for formatting energy data from PyRETIS.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

EnergyFormatter (:py:class:`.EnergyFormatter`)
    A class for formatting energy data from PyRETIS.

EnergyPathFormatter (:py:class:`.EnergyPathFormatter`)
    A class for formatting energy data for paths.

EnergyFile (:py:class:`.EnergyFile`)
    A class for handling PyRETIS energy files.

EnergyPathFile (:py:class:`.EnergyPathFile`)
    A class for handling PyRETIS energy path files.

"""
import logging
import numpy as np
from pyretis.inout.formats.formatter import OutputFormatter
from pyretis.inout.fileio import FileIO, read_some_lines
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

np.set_printoptions(legacy='1.21')

__all__ = [
    'EnergyFormatter',
    'EnergyPathFormatter',
    'EnergyFile',
    'EnergyPathFile',
]


class EnergyFormatter(OutputFormatter):
    """A class for formatting energy data from PyRETIS.

    This class handles formatting of energy data.
    The data is formatted in 5 columns:

    1) Time, i.e. the step number.

    2) Potential energy.

    3) Kinetic energy.

    4) Total energy, should equal the sum of the two previous columns.

    5) Temperature.

    """

    # Format for the energy files:
    ENERGY_FMT = ['{:>10d}'] + 5*['{:>14.6f}']
    ENERGY_TERMS = ('vpot', 'ekin', 'etot', 'temp')
    HEADER = {'labels': ['Time', 'Potential', 'Kinetic', 'Total',
                         'Temperature'],
              'width': [10, 14]}

    def __init__(self, name='EnergyFormatter'):
        """Initialise the formatter for energy."""
        super().__init__(name, header=self.HEADER)

    def apply_format(self, step, energy):
        """Apply the energy format.

        Parameters
        ----------
        step : int
            The current simulation step.
        energy : dict
            A dict with energy terms to format.

        Returns
        -------
        out : string
            A string with the formatted energy data.

        """
        towrite = [self.ENERGY_FMT[0].format(step)]
        for i, key in enumerate(self.ENERGY_TERMS):
            value = energy.get(key, None)
            if value is None:
                towrite.append(self.ENERGY_FMT[i + 1].format(float('nan')))
            else:
                towrite.append(self.ENERGY_FMT[i + 1].format(float(value)))
        return ' '.join(towrite)

    def format(self, step, data):
        """Yield formatted energy data. See :py:meth:.`apply_format`."""
        yield self.apply_format(step, data)

    def load(self, filename):
        """Load entire energy blocks into memory.

        Parameters
        ----------
        filename : string
            The path/file name of the file we want to open.

        Yields
        ------
        data_dict : dict
            This is the energy data read from the file, stored in
            a dict. This is for convenience so that each energy term
            can be accessed by `data_dict['data'][key]`.

        """
        for blocks in read_some_lines(filename, line_parser=self.parse):
            data = np.array(blocks['data'])
            col = tuple(data.shape)
            col_max = min(col[1], len(self.ENERGY_TERMS) + 1)
            data_dict = {'comment': blocks['comment'],
                         'data': {'time': data[:, 0]}}
            for i in range(col_max-1):
                data_dict['data'][self.ENERGY_TERMS[i]] = data[:, i+1]
            yield data_dict


class EnergyPathFormatter(EnergyFormatter):
    """A class for formatting energy data for paths."""

    ENERGY_TERMS = ('vpot', 'ekin')
    HEADER = {'labels': ['Time', 'Potential', 'Kinetic'],
              'width': [10, 14]}

    def __init__(self):
        """Initialise."""
        super().__init__(name='EnergyPathFormatter')
        self.print_header = False

    def format(self, step, data):
        """Format the order parameter data from a path.

        Parameters
        ----------
        step : int
            The cycle number we are creating output for.
        data : tuple
            Here we assume that ``data[0]`` contains an object
            like :py:class:`.PathBase` and that ``data[1]`` is a
            string with the status for the path.

        Yields
        ------
        out : string
            The strings to be written.

        """
        path, status = data[0], data[1]
        if not path:  # when nullmoves = False
            return
        move = path.generated
        yield f'# Cycle: {step}, status: {status}, move: {move}'
        yield self.header
        for i, phasepoint in enumerate(path.phasepoints):
            energy = {}
            for key in self.ENERGY_TERMS:
                energy[key] = getattr(phasepoint.particles, key, None)
            yield self.apply_format(i, energy)


class EnergyFile(FileIO):
    """A class for handling PyRETIS energy files."""

    def __init__(self, filename, file_mode, backup=True):
        """Create the file object and attach the energy formatter."""
        super().__init__(filename, file_mode, EnergyFormatter(), backup=backup)


class EnergyPathFile(FileIO):
    """A class for handling PyRETIS energy path files."""

    def __init__(self, filename, file_mode, backup=True):
        """Create the file object and attach the energy formatter."""
        super().__init__(filename, file_mode, EnergyPathFormatter(),
                         backup=backup)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definitions of simulation objects for molecular dynamics simulations.

This module contains definitions of classes for performing molecular
dynamics simulations.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SimulationNVE (:py:class:`.SimulationNVE`)
    Definition of a simple NVE simulation. The engine
    used for this simulation must have dynamics equal to NVE.

SimulationMD (:py:class:`.SimulationMD`)
    Definition of a simulation for running somply MD.

SimulationMDFlux (:py:class:`.SimulationMDFlux`)
    Definition of a simulation for determining the initial flux.
    This is used for calculating rates in TIS simulations.
"""
import logging
from pyretis.simulation.simulation import Simulation
from pyretis.core.particlefunctions import calculate_thermo
from pyretis.core.path import check_crossing
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = [
    'SimulationMD',
    'SimulationNVE',
    'SimulationMDFlux'
]


class SimulationMD(Simulation):
    """A generic MD simulation.

    This class is used to define a MD simulation without whistles and bells.

    """

    simulation_type = 'md'
    simulation_output = [
        {'type': 'energy', 'name': 'md-energy-file'},
        {'type': 'thermo-file', 'name': 'md-thermo-file'},
        {'type': 'traj-xyz', 'name': 'md-traj-file'},
        {'type': 'thermo-screen', 'name': 'md-thermo-screen'},
        {'type': 'order', 'name': 'md-order-file'},
    ]

    def __init__(self, ensemble, settings=None, controls=None):
        """Only add variable.

        Parameters
        ----------
        ensemble : dict
            It contains the simulations info

            * `system` : object like :py:class:`.System`
              This is the system we are investigating.
            * `engine` : object like :py:class:`.EngineBase`
              This is the integrator that is used to propagate
              the system in time.
            * `order_function` : object like :py:class:`.OrderParameter`,
              optional. A class that can be used to calculate an order
              parameter, if needed.

        settings : dict, optional
            This dictionary contains the settings for the simulation.
        controls: dict of parameters, optional
            It contains:

            * `steps` : int, optional
              The number of simulation steps to perform.
            * `startcycle` : int, optional
              The cycle we start the simulation on, can be useful if
              restarting.

        """
        super().__init__(settings, controls)
        self.engine = ensemble['engine']
        self.system = ensemble['system']
        self.order_function = ensemble.get('order_function')
        self.system.potential_and_force()  # make sure forces are defined.
        self.ensemble = ensemble

    def run(self):
        """Run the MD simulation.

        Yields
        ------
        results : dict
            The results from a single step in the simulation.

        """
        nsteps = 1 + self.cycle['endcycle'] - self.cycle['step']
        integ = self.engine.integrate({'system': self.system,
                                       'order_function': self.order_function},
                                      nsteps,
                                      thermo='full')
        for step in integ:
            if not self.first_step:
                self.cycle['step'] += 1
                self.cycle['stepno'] += 1
            results = {'cycle': self.cycle.copy()}
            results.update(self.execute_tasks())
            if self.first_step:
                self.first_step = False
            results.update(step)
            for task in self.output_tasks:
                task.output(results)
            self.write_restart()
            if self.soft_exit():
                yield results
                break
            yield results

    def __str__(self):
        """Return a string with info about the simulation."""
        msg = ['MD simulation']
        nstep = self.cycle['endcycle'] - self.cycle['startcycle']
        msg += [f'Number of steps to do: {nstep}']
        msg += [f'Molecular Dynamics engine: {self.engine}']
        msg += [f'Time step: {self.engine.timestep}']
        return '\n'.join(msg)


class SimulationNVE(SimulationMD):
    """A MD NVE simulation class.

    This class is used to define a NVE simulation. Compared with
    the :py:class:`.SimulationMD` we here require that the engine
    supports NVE dynamics.
    """

    simulation_type = 'md-nve'
    simulation_output = [
        {'type': 'energy', 'name': 'nve-energy-file'},
        {'type': 'thermo-file', 'name': 'nve-thermo-file'},
        {'type': 'traj-xyz', 'name': 'nve-traj-file'},
        {'type': 'thermo-screen', 'name': 'nve-thermo-screen'},
        {'type': 'order', 'name': 'nve-order-file'},
    ]

    def __init__(self, ensemble, settings=None, controls=None):
        """Initialise the NVE simulation object.

        Here we will set up the tasks that are to be performed in the
        simulation, such as the integration and thermodynamics
        calculation(s).

        Parameters
        ----------
        ensemble : dict
            It contains the simulations info

            * `system` : object like :py:class:`.System`
              This is the system we are investigating.
            * `engine` : object like :py:class:`.EngineBase`
              This is the integrator that is used to propagate the
              system in time.
            * `order_function` : object like :py:class:`.OrderParameter`, opt
              A class that can be used to calculate an order parameter,
              if needed.

        settings : dict, optional
            This dictionary contains the settings for the simulation.
        controls: dict of parameters, optional
            It contains:

            * `steps` : int, optional
              The number of simulation steps to perform.
            * `startcycle` : int, optional
              The cycle we start the simulation on.

        """
        super().__init__(ensemble, settings, controls)
        dyn = getattr(self.engine, "dynamics", "")
        if dyn.lower() != 'nve':
            logger.warning(
                'MD integrator %s (%s) might not be suitable'
                ' for NVE dynamics! Please make sure this is intentional.',
                self.engine.__class__,
                self.engine.description
            )

    def step(self):
        """Run a single simulation step."""
        system = self.ensemble['system']
        if self.first_step:
            system.potential_and_force()
            self.first_step = False
        else:
            self.cycle['step'] += 1
            self.cycle['stepno'] += 1
            self.engine.integration_step(system)
        results = {'cycle': self.cycle.copy(),
                   'thermo': calculate_thermo(system),
                   'system': system}
        if self.order_function:
            results['order'] = self.engine.calculate_order(self.ensemble)
        return results

    def __str__(self):
        """Return a string with info about the simulation."""
        msg = ['NVE simulation']
        nstep = self.cycle['endcycle'] - self.cycle['startcycle']
        msg += [f'Number of steps to do: {nstep}']
        msg += [f'MD engine: {self.engine}']
        msg += [f'Time step: {self.engine.timestep}']
        return '\n'.join(msg)

    def restart_info(self):
        """Return information which can be used to restart the simulation.

        Returns
        -------
        info : dict,
            Contains all the updated simulation settings and counters.

        """
        info = super().restart_info()
        info['engine'] = self.engine.restart_info()

        return info

    def load_restart_info(self, info):
        """Load the restart information."""
        super().load_restart_info(info)
        self.engine.load_restart_info(info['engine'])


class SimulationMDFlux(SimulationMD):
    """A simulation for obtaining the initial flux for TIS.

    This class is used to define a MD simulation where the goal is
    to calculate crossings in order to obtain the initial flux for a TIS
    calculation.

    """

    simulation_type = 'md-flux'
    simulation_output = [
        {'type': 'energy', 'name': 'flux-energy-file'},
        {'type': 'traj-xyz', 'name': 'flux-traj-file'},
        {'type': 'thermo-screen', 'name': 'flux-thermo-screen'},
        {'type': 'order', 'name': 'flux-order-file'},
        {'type': 'cross', 'name': 'flux-cross-file'},
    ]

    def __init__(self, ensemble, settings=None, controls=None):
        """Initialise the MD-Flux simulation object.

        Parameters
        ----------
        ensemble : dict
            It contains the simulations info

            * `system` : object like :py:class:`.System`
              The system to act on.
            * `engine` : object like :py:class:`.EngineBase`
              This is the integrator that is used to propagate the system
              in time.
            * `order_function` : object like :py:class:`.OrderParameter`
              The class used for calculating the order parameters.

        settings : dict, optional
            This dictionary contains the settings for the simulation.
        controls: dict of parameters, optional
            It contains:

            * `steps` : int, optional
              The number of simulation steps to perform.
            * `startcycle` : int, optional
              The cycle we start the simulation on, can be useful if
              restarting.

        """
        super().__init__(ensemble, settings, controls)

        # set up for initial crossing
        self.leftside_prev = None
        self.interfaces = settings['simulation'].get('interfaces')

    def run(self):
        """Run the MD simulation.

        Yields
        ------
        results : dict
            The results from a single step in the simulation.

        """
        nsteps = 1 + self.cycle['endcycle'] - self.cycle['step']
        leftside = None
        integ = self.engine.integrate(
            {'system': self.system,
             'order_function': self.order_function},
            nsteps,
            thermo='full',
        )
        for step in integ:
            if not self.first_step:
                self.cycle['step'] += 1
                self.cycle['stepno'] += 1
            else:
                self.first_step = False
            results = {'cycle': self.cycle.copy()}
            if leftside:
                self.leftside_prev = leftside
            leftside, cross = check_crossing(self.cycle['step'],
                                             step['order'][0],
                                             self.interfaces,
                                             self.leftside_prev)
            results['cross'] = cross
            results.update(step)

            for task in self.output_tasks:
                task.output(results)
            self.write_restart()
            if self.soft_exit():
                yield results
                break
            yield results

    def __str__(self):
        """Return a string with info about the simulation."""
        msg = ['MD-flux simulation']
        nstep = self.cycle['endcycle'] - self.cycle['startcycle']
        msg += [f'Number of steps to do: {nstep}']
        msg += [f'Molecular Dynamics engine: {self.engine}']
        msg += [f'Time step: {self.engine.timestep}']
        return '\n'.join(msg)

    def restart_info(self):
        """Return information which can be used to restart the simulation.

        Returns
        -------
        info : dict
            Contains all the updated simulation settings and counters.

        """
        info = super().restart_info()
        info['engine'] = self.engine.restart_info()
        info['leftside_prev'] = self.leftside_prev
        return info

    def load_restart_info(self, info):
        """Load the restart information."""
        super().load_restart_info(info)
        self.leftside_prev = info['leftside_prev']
        self.engine.load_restart_info(info['engine'])



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of the OpenMM engine.

This module defines the class for the OpenMM MD engine.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OpenMMEngine (:py:class:`.OpenMMEngine`)
    The class for running the OpenMM engine.

"""

import logging
from pyretis.core import System, Particles
from pyretis.core.particlefunctions import (calculate_kinetic_energy,
                                            reset_momentum)
from pyretis.engines.engine import EngineBase
from pyretis.core.box import create_box, box_matrix_to_list
from pyretis.core.common import import_from
try:
    import openmm
    import openmm.unit as u
    HAS_OPENMM = True
except ImportError:
    try:
        from simtk import openmm
        import simtk.unit as u  # pragma: no cover
        HAS_OPENMM = True  # pragma: no cover
    except ImportError:  # pragma: no cover
        HAS_OPENMM = False  # pragma: no cover

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

__all__ = ['OpenMMEngine']


def make_pyretis_system(settings):
    """Make a PyRETIS system from OpenMM Simulation.

    First this looks for setting->engine->openmm_module and openmm_simulation.
    It then imports that openmm_simulation object and creates a internal
    pyretis system representation
    """
    simulation = import_from(settings['engine']['openmm_module'],
                             settings['engine']['openmm_simulation'])
    state = simulation.context.getState(getPositions=True, getVelocities=True)
    box = state.getPeriodicBoxVectors(asNumpy=True).T
    box = create_box(cell=box_matrix_to_list(box),
                     periodic=[True, True, True])
    system = System(units='gromacs', box=box,
                    temperature=simulation.integrator.getTemperature())
    system.particles = Particles(system.get_dim())
    pos = state.getPositions(asNumpy=True)
    vel = state.getVelocities(asNumpy=True)
    for i, atom in enumerate(simulation.topology.atoms()):
        mass = simulation.system.getParticleMass(i)
        mass = mass.value_in_unit(u.grams/u.mole)
        name = atom.name
        system.particles.add_particle(pos=pos[i], vel=vel[i],
                                      force=[None, None, None],
                                      mass=mass, name=name)
    return system


class OpenMMEngine(EngineBase):
    """
    A class for interfacing with OpenMM.

    This class defines the interface to OpenMM.

    Attributes
    ----------
    simulation: OpenMM.simulation
        OpenMM simulation object.

    subcyles: int
        Number of OpenMM steps per PyRETIS step.

    """

    engine_type = 'openmm'

    def __init__(self, openmm_simulation, subcycles=1, openmm_module=None):
        """Set up the OpenMM Engine.

        Parameters
        ----------
        openmm_simulation : OpenMM.simulation, or string
            The OpenMM simulation object or the variable name if `module` is
            not None.

        subcycles: int
            Number of OpenMM integration steps per PyRETIS step.

        openmm_module: string, optional
            If defined and simulation is a string, try loading `simulation`
            from `module`.

        """
        # Check if openmm is installed.
        if HAS_OPENMM is False:
            raise RuntimeError("OpenMM is not installed")
        super(OpenMMEngine, self).__init__(description='OpenMM')
        if isinstance(openmm_simulation, str) and openmm_module is not None:
            openmm_simulation = import_from(openmm_module, openmm_simulation)
        if isinstance(openmm_simulation, str):
            msg = "simulation is a string but could not be loaded from module"
            msg += f" '{openmm_module}'"
            raise RuntimeError(msg)

        self.simulation = openmm_simulation
        self.subcycles = subcycles

    def calculate_order(self, ensemble, xyz=None, vel=None, box=None):
        """Return the order parameter.

        This method is just to help to calculate the order parameter
        in cases where only the engine can do it.

        Parameters
        ----------
        ensemble : dict
            It contains:

            * `system` : object like :py:class:`.System`
              This is the system that contains the particles we are
              investigating
            * `order_function` : object like :py:class:`.OrderParameter`
              The class used for calculating the order parameter.
        xyz : numpy.array, optional
            The positions to use. Typically for internal engines, this
            is not needed. It is included here as it can be used for
            testing and also to be compatible with the generic function
            defined by the parent.
        vel : numpy.array, optional
            The velocities to use.
        box : numpy.array, optional
            The current box vectors.

        Returns
        -------
        out : list of floats
            The calculated order parameter(s).

        """
        system = ensemble['system']
        order_function = ensemble['order_function']
        if any((xyz is None, vel is None, box is None)):
            return order_function.calculate(system)
        system.particles.pos = xyz
        system.particles.vel = vel
        system.box.update_size(box)
        return order_function.calculate(system)

    def clean_up(self):
        """Clean up after using the engine.

        Currently, this is only included for compatibility with external
        integrators.

        """
        return

    def dump_phasepoint(self, phasepoint, deffnm=None):
        """For compatibility with external integrators."""
        return

    def integration_step(self, system):
        """Perform one integration step of n subcycles.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            system.

        """
        particles = system.particles
        pos = particles.pos
        vel = particles.vel
        box = system.box
        context = self.simulation.context

        # Update the simulation to the correct system:
        context.setPositions(pos)
        context.setVelocities(vel)

        # PyRETIS matrix.T is openmm boxvectors:
        if box is not None:
            vectors = list(box.box_matrix.T)
            context.setPeriodicBoxVectors(*vectors)

        # Run the simulation for n subcycles:
        self.simulation.step(self.subcycles)

        # Get the current state:
        state = context.getState(getPositions=True, getVelocities=True)

        # Update system:
        system.particles.pos = state.getPositions(asNumpy=True)
        system.particles.vel = state.getVelocities(asNumpy=True)
        if box is not None:
            box.update_size(
                box_matrix_to_list(state.getPeriodicBoxVectors(asNumpy=True).T)
            )

    def kick_across_middle(self, ensemble, middle, tis_settings):
        """Force a phase point across the middle interface.

        This is accomplished by repeatedly kicking the phase point so
        that it crosses the middle interface.

        Parameters
        ----------
        ensemble : dict
            It contains:

            * `system` : object like :py:class:`.System`
              This is the system that contains the particles we are
              investigating
            * `order_function` : object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.
            * `rgen` : object like :py:class:`.RandomGenerator`
              This is the random generator that will be used.
        middle : float
            This is the value for the middle interface.
        tis_settings : dict
            This dictionary contains settings for TIS. Explicitly used here:

            * `zero_momentum`: boolean, determines if the momentum is zeroed.
            * `rescale_energy`: boolean, determines if energy is re-scaled.


        Returns
        -------
        out[0] : object like :py:class:`.System`
            The phase-point just before the interface.
        out[1] : object like :py:class:`.System`
            The phase-point just after the interface.

        Note
        ----
        This function will update the system state.

        """
        # Taken from MDEngine in internal.py
        # We search for crossing with the middle interface and do this
        # by sequentially kicking the initial phase point:
        system = ensemble['system']
        previous = None
        curr = self.calculate_order(ensemble)[0]
        while True:  # pragma: no cover
            # Save current state:
            previous = system.copy()
            # Save previous box matrix:
            prev_box = system.box.box_matrix
            # Modify velocities:
            vel_settings = {'sigma_v': None, 'aimless': True,
                            'momentum': tis_settings['zero_momentum'],
                            'rescale': tis_settings['rescale_energy']}
            self.modify_velocities(ensemble, vel_settings)
            # Update order parameter in case it is velocity dependent:
            curr = self.calculate_order(ensemble)[0]
            previous.order = curr
            # Store modified velocities:
            previous.particles.set_vel(system.particles.get_vel())
            # Integrate forward one step:
            self.integration_step(system)
            # Compare previous order parameter and the new one:
            prev = curr
            curr = self.calculate_order(ensemble)[0]
            if (prev < middle < curr) or (curr < middle < prev):
                # Middle interface was crossed, just stop the loop.
                logger.info('Crossing found: %9.6f %9.6f ', prev, curr)
                break
            if (prev <= curr <= middle) or (middle <= curr <= prev):
                # We are getting closer, keep the new point.
                pass
            elif prev == middle:
                # Unlucky case, we want more than 1 point, so search more.
                pass
            else:  # We did not get closer, fall back to previous point.
                system.particles = previous.particles.copy()
                system.box.update_size(box_matrix_to_list(prev_box))
                curr = previous.order
        return previous, system

    def modify_velocities(self, ensemble, vel_settings):
        """Modify the velocities of the current state.

        This method will modify the velocities of a time slice.
        And it is part of the integrator since it, conceptually,
        fits here:  we are acting on the system and modifying it.

        Parameters
        ----------
        ensemble : dict
            It contains:

            * `system` : object like :py:class:`.System`
              This is the system that contains the particles we are
              investigating
            * `rgen` : object like :py:class:`.RandomGenerator`
              This is the random generator that will be used.

        vel_settings: dict.
            It contains info about velocity settings:

            * `sigma_v` : numpy.array, optional
              These values can be used to set a standard deviation (one
              for each particle) for the generated velocities.
            * `aimless` : boolean, optional
              Determines if we should do aimless shooting or not.
            * `momentum` : boolean, optional
              If True, we reset the linear momentum to zero after
              generating.
            * `rescale` : float, optional
              In some NVE simulations, we may wish to re-scale the
              energy to a fixed value. If `rescale` is a float > 0,
              we will re-scale the energy (after modification of
              the velocities) to match the given float.

        Returns
        -------
        dek : float
            The change in the kinetic energy.
        kin_new : float
            The new kinetic energy.

        """
        system = ensemble['system']
        rescale = vel_settings.get('rescale')
        particles = system.particles
        rgen = ensemble['rgen']
        if rescale is not None and rescale is not False:
            if rescale > 0:
                kin_old = rescale - particles.vpot
                do_rescale = True
            else:
                logger.warning('Ignored re-scale 6.2%f < 0.0.', rescale)
                return 0.0, calculate_kinetic_energy(particles)[0]
        else:
            kin_old = calculate_kinetic_energy(particles)[0]
            do_rescale = False
        if vel_settings.get('aimless', False):
            vel, _ = rgen.draw_maxwellian_velocities(system=system)
            particles.vel = vel
        else:  # Soft velocity change, from a Gaussian distribution:
            dvel, _ = rgen.draw_maxwellian_velocities(
                system=system,
                sigma_v=vel_settings['sigma_v'])
            particles.vel = particles.vel + dvel
        if vel_settings.get('momentum', False):
            reset_momentum(particles)
        if do_rescale:
            system.rescale_velocities(rescale)
        kin_new = calculate_kinetic_energy(particles)[0]
        dek = kin_new - kin_old
        return dek, kin_new

    def propagate(self, path, ensemble, reverse=False):
        """Generate a path by integrating until a criterion is met.

        This function will generate a path by calling the function
        specifying the integration step repeatedly. The integration is
        carried out until the order parameter has passed the specified
        interfaces or if we have integrated for more than a specified
        maximum number of steps. The given system defines the initial
        state and the system is reset to its initial state when this
        method is done.

        Parameters
        ----------
        path : object like :py:class:`.PathBase`
            This is the path we use to fill in phase-space points.
            We are here not returning a new path - this since we want
            to delegate the creation of the path (type) to the method
            that is running `propagate`.
        ensemble: dict
            it contains:

            * `system` : object like :py:class:`.System`
              The system object gives the initial state for the
              integration. The initial state is stored and the system is
              reset to the initial state when the integration is done.
            * `order_function` : object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.
            * `interfaces` : list of floats
              These interfaces define the stopping criterion.

        reverse : boolean, optional
            If True, the system will be propagated backward in time.

        Returns
        -------
        success : boolean
            This is True if we generated an acceptable path.
        status : string
            A text description of the current status of the propagation.

        """
        system = ensemble['system']
        status = 'Propagate w/OpenMM engine'
        logger.debug(status)
        success = False
        left, _, right = ensemble['interfaces']
        for i in range(path.maxlen):
            order = self.calculate_order(ensemble)
            kin = calculate_kinetic_energy(system.particles)[0]
            snapshot = {'order': order,
                        'pos': system.particles.pos,
                        'vel': system.particles.vel,
                        'box': system.box,
                        'ekin': kin,
                        'vpot': None}
            phasepoint = self.snapshot_to_system(system, snapshot)
            status, success, stop, _ = self.add_to_path(path, phasepoint,
                                                        left, right)
            if stop:
                logger.debug('Stopping propagate at step: %i ', i)
                break

            if reverse:
                system.particles.vel *= -1.0
                self.integration_step(system)
                system.particles.vel *= -1.0
            else:
                self.integration_step(system)
        logger.debug('Propagate done: "%s" (success: %s)', status, success)
        return success, status
